/******************************************************************************
 * sdla_56k.c	WANPIPE(tm) Multiprotocol WAN Link Driver. 
 *				56K board configuration.
 *
 * Author: 	Nenad Corbic  <ncorbic@sangoma.com>
 *
 * Copyright:	(c) 1995-2001 Sangoma Technologies Inc.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 * ============================================================================
 * Jul 22, 2001	Nenad Corbic	Initial version.
 * Oct 01, 2001 Gideon Hack	Modifications for interrupt usage.
 ******************************************************************************
 */


/*
 ******************************************************************************
			   INCLUDE FILES
 ******************************************************************************
*/

#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
# include <net/wanpipe_config.h>
# include <net/wanpipe_includes.h>
# include <net/wanpipe.h>
#elif defined(__WINDOWS__)
# include <wanpipe\wanpipe_includes.h>
# include <hdlc_stm_api\driver.h>
#else
# include <linux/wanpipe_includes.h>
# include <linux/wanpipe.h>	/* WANPIPE common user API definitions */
#endif

/*
 ******************************************************************************
			  DEFINES AND MACROS
 ******************************************************************************
*/
#if 0
#define WRITE_REG(reg, value)		card->wandev.write_front_end_reg(card, reg, (unsigned char)(value))
#define READ_REG(reg)			card->wandev.read_front_end_reg(card, reg)
#endif

/*
 ******************************************************************************
			STRUCTURES AND TYPEDEFS
 ******************************************************************************
*/


/*
 ******************************************************************************
			   GLOBAL VARIABLES
 ******************************************************************************
*/

/*
 ******************************************************************************
			  FUNCTION PROTOTYPES
 ******************************************************************************
*/
static void display_Rx_code_condition(sdla_t* card);
static unsigned long sdla_56k_service_status(void* card_id);
static void sdla_56k_alarm_print(void* card_id, unsigned long);

/*
 ******************************************************************************
			  FUNCTION DEFINITIONS
 ******************************************************************************
*/
unsigned short sdla_56k_alarm(void* arg1, char manual_read)
{

	unsigned short status = 0x00;
	sdla_t* card = (sdla_t *)arg1;
	wan_device_t* wandev = &card->wandev; 

	WAN_ASSERT(wandev->write_front_end_reg == NULL);
	WAN_ASSERT(wandev->read_front_end_reg == NULL);
	/* set to manually read the front-end registers */
	if(manual_read) {
		wandev->RR8_reg_56k = 0;
		wandev->RR8_reg_56k = READ_REG(REG_INT_EN_STAT);
		wandev->RRA_reg_56k = READ_REG(REG_DEV_STAT);
		wandev->RRC_reg_56k = READ_REG(REG_RX_CODES);
		status |= wandev->RRA_reg_56k;
		status = (status << 8) & 0xFF00;
		status |= wandev->RRC_reg_56k;
	}
	
	DEBUG_56K("56K registers: 8-%02X,C-%02X,A-%02X\n", 
			wandev->RR8_reg_56k,
			wandev->RRC_reg_56k, 
			wandev->RRA_reg_56k);

	/* discard an invalid Rx code interrupt */
	if((wandev->RR8_reg_56k & BIT_INT_EN_STAT_RX_CODE) && 
		(!wandev->RRC_reg_56k)) {
		return(status);
	}

	/* record an RLOS (Receive Loss of Signal) condition */
	if(wandev->RRA_reg_56k & BIT_DEV_STAT_RLOS) {
		wandev->RLOS_56k = 1;
	}

	/* display Rx code conditions if necessary */
	if(!((wandev->RRC_reg_56k ^ wandev->prev_RRC_reg_56k) & 
		wandev->delta_RRC_reg_56k) ||
		(wandev->front_end_status == FE_CONNECTED)) {
			display_Rx_code_condition(card);
	}
	
	/* record the changes that occurred in the Rx code conditions */
	wandev->delta_RRC_reg_56k |= 
		(wandev->RRC_reg_56k ^ wandev->prev_RRC_reg_56k);

	
	if(manual_read || (wandev->RR8_reg_56k & BIT_INT_EN_STAT_ACTIVE)) {

		/* if Insertion Loss is less than 44.4 dB, then we are connected */
		if ((wandev->RRA_reg_56k & 0x0F) > BIT_DEV_STAT_IL_44_dB) {
			if((wandev->front_end_status == FE_DISCONNECTED) ||
			 (wandev->front_end_status == FE_UNITIALIZED)) {
				
				wandev->front_end_status = FE_CONNECTED;
				/* reset the Rx code condition changes */
				wandev->delta_RRC_reg_56k = 0;
				
				if(wandev->RLOS_56k == 1) {
					wandev->RLOS_56k = 0;
					DEBUG_EVENT("%s: 56k Receive Signal Recovered\n", 
							card->devname);
				}
				DEBUG_EVENT("%s: 56k Connected\n",
							card->devname);
			}
		}else{
			if((wandev->front_end_status == FE_CONNECTED) || 
			 (wandev->front_end_status == FE_UNITIALIZED)) {
				
				wandev->front_end_status = FE_DISCONNECTED;
				/* reset the Rx code condition changes */
				wandev->delta_RRC_reg_56k = 0;

				if(wandev->RLOS_56k == 1) {
					DEBUG_EVENT("%s: 56k Receive Loss of Signal\n", 
							card->devname);
				}
				DEBUG_EVENT("%s: 56k Disconnected (loopback)\n", 
						card->devname);
			}
		}
	}
	
	wandev->prev_RRC_reg_56k = wandev->RRC_reg_56k;
	return(status);
}


int sdla_56k_default_cfg(void* pcard, void* p56k_cfg)
{
	return 0;
}


int sdla_56k_config(void* arg1)
{
	sdla_t* card = (sdla_t *)arg1;
	wan_device_t* wandev = &card->wandev; 

	WAN_ASSERT(wandev->write_front_end_reg == NULL);
	WAN_ASSERT(wandev->read_front_end_reg == NULL);
	/* The 56k CSU/DSU front end status has not been initialized  */
	wandev->front_end_status = FE_UNITIALIZED;

	/* Zero the previously read RRC register value */
	wandev->prev_RRC_reg_56k = 0;
	
	/* Zero the RRC register changes */ 
	wandev->delta_RRC_reg_56k = 0;
	
	if(WRITE_REG(REG_INT_EN_STAT, (BIT_INT_EN_STAT_IDEL | 
		BIT_INT_EN_STAT_RX_CODE | BIT_INT_EN_STAT_ACTIVE))) {
		return 1;
	} 

	if(WRITE_REG(REG_RX_CODES, 0xFF & ~(BIT_RX_CODES_ZSC))) {
		return 1;
	}

	if (card->wandev.clocking == WANOPT_EXTERNAL){
		DEBUG_EVENT("%s: Configuring 56K CSU/DSU for External Clocking\n",
				card->devname);
		if(WRITE_REG(REG_DEV_CTRL, 
			(BIT_DEV_CTRL_SCT_E_OUT | BIT_DEV_CTRL_DDS_PRI))) {
			return 1;
		}
	}else{
		DEBUG_EVENT("%s: Configuring 56K CSU/DSU for Internal Clocking\n",
				card->devname);
		if(WRITE_REG(REG_DEV_CTRL, (BIT_DEV_CTRL_XTALI_INT | 
			BIT_DEV_CTRL_SCT_E_OUT | BIT_DEV_CTRL_DDS_PRI))) {
			return 1;
		}
 	}

	if(WRITE_REG(REG_TX_CTRL, 0x00)) {
		return 1;
	}

	if(WRITE_REG(REG_RX_CTRL, 0x00)) {
		return 1;
	}

	if(WRITE_REG(REG_EIA_SEL, 0x00)) {
		return 1;
	}

	if(WRITE_REG(REG_EIA_TX_DATA, 0x00)) {
		return 1;
	}

	if(WRITE_REG(REG_EIA_CTRL, (BIT_EIA_CTRL_RTS_ACTIVE | 
		BIT_EIA_CTRL_DTR_ACTIVE | BIT_EIA_CTRL_DTE_ENABLE))) {
		return 1; 
	}

	/* Initialize function pointer for WANPIPE debugging */
	card->fe_iface.get_fe_service_status = &sdla_56k_service_status;
	card->fe_iface.print_fe_alarm = &sdla_56k_alarm_print;
	return 0;
}


static void display_Rx_code_condition(sdla_t* card)
{
	wan_device_t* wandev = &card->wandev;

	/* check for a DLP (DSU Non-latching loopback) code condition */
	if((wandev->RRC_reg_56k ^ wandev->prev_RRC_reg_56k) &
		BIT_RX_CODES_DLP) {
		if(wandev->RRC_reg_56k & BIT_RX_CODES_DLP) {
			DEBUG_EVENT("%s: 56k receiving DSU Loopback code\n", 
					card->devname);
		} else {
			DEBUG_EVENT("%s: 56k DSU Loopback code condition ceased\n", 
					card->devname);
		}
	}

	/* check for an OOF (Out of Frame) code condition */
	if((wandev->RRC_reg_56k ^ wandev->prev_RRC_reg_56k) &
		BIT_RX_CODES_OOF) {
		if(wandev->RRC_reg_56k & BIT_RX_CODES_OOF) {
			DEBUG_EVENT("%s: 56k receiving Out of Frame code\n", 
					card->devname);
		} else {
			DEBUG_EVENT("%s: 56k Out of Frame code condition ceased\n", 
					card->devname);
		}
	}

	/* check for an OOS (Out of Service) code condition */
	if((wandev->RRC_reg_56k ^ wandev->prev_RRC_reg_56k) &
		BIT_RX_CODES_OOS) {
		if(wandev->RRC_reg_56k & BIT_RX_CODES_OOS) {
			DEBUG_EVENT("%s: 56k receiving Out of Service code\n", 
					card->devname);
		} else {
			DEBUG_EVENT("%s: 56k Out of Service code condition ceased\n", 
					card->devname);
		}
	}

	/* check for a CMI (Control Mode Idle) condition */
	if((wandev->RRC_reg_56k ^ wandev->prev_RRC_reg_56k) &
		BIT_RX_CODES_CMI) {
		if(wandev->RRC_reg_56k & BIT_RX_CODES_CMI) {
			DEBUG_EVENT("%s: 56k receiving Control Mode Idle\n",
					card->devname);
		} else {
			DEBUG_EVENT("%s: 56k Control Mode Idle condition ceased\n", 
					card->devname);
		}
	}

	/* check for a ZSC (Zero Suppression Code) condition */
	if((wandev->RRC_reg_56k ^ wandev->prev_RRC_reg_56k) &
		BIT_RX_CODES_ZSC) {
		if(wandev->RRC_reg_56k & BIT_RX_CODES_ZSC) {
			DEBUG_EVENT("%s: 56k receiving Zero Suppression code\n", 
					card->devname);
		} else {
			DEBUG_EVENT("%s: 56k Zero Suppression code condition ceased\n", 
					card->devname);
		}
	}

	/* check for a DMI (Data Mode Idle) condition */
	if((wandev->RRC_reg_56k ^ wandev->prev_RRC_reg_56k) &
		BIT_RX_CODES_DMI) {
		if(wandev->RRC_reg_56k & BIT_RX_CODES_DMI) {
			DEBUG_EVENT("%s: 56k receiving Data Mode Idle\n",
				card->devname);
		} else {
			DEBUG_EVENT("%s: 56k Data Mode Idle condition ceased\n", 
				card->devname);
		}
	}
}

/*
 ******************************************************************************
 *				sdla_56k_udp()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
int sdla_56k_udp(void* card_id, wan_cmd_t* udp_cmd, unsigned char* data)
{
	sdla_t *card = (sdla_t*)card_id;

	switch(udp_cmd->wan_cmd_command){
	case WAN_GET_MEDIA_TYPE:
		data[0] = (IS_56K_CARD(card) ? WANOPT_MEDIA_56K : 
						WANOPT_MEDIA_NONE);
		udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
		udp_cmd->wan_cmd_data_len = sizeof(unsigned char); 
		break;

	case WAN_FE_GET_STAT:
		/* 56K Update CSU/DSU alarms */
		card->wandev.k56_alarm = sdla_56k_alarm(card, 1); 
	 	*(unsigned long *)&data[0] = card->wandev.k56_alarm;
		udp_cmd->wan_cmd_return_code = WAN_CMD_OK;
	    	udp_cmd->wan_cmd_data_len = sizeof(unsigned long);
		break;

	case WAN_FE_SET_LB_MODE:
 	case WAN_FE_FLUSH_PMON:
	case WAN_FE_GET_CFG:
	default:
		udp_cmd->wan_cmd_return_code = WAN_UDP_INVALID_CMD;
	    	udp_cmd->wan_cmd_data_len = 0;
		break;
	}
	return 0;
}

/*
 ******************************************************************************
 *				sdla_56k_service_status()	
 *
 * Description:
 * Arguments:
 * Returns:
 ******************************************************************************
 */
static unsigned long sdla_56k_service_status(void* card_id)
{
	return (unsigned long)sdla_56k_alarm(card_id, 0);
}

/*
*******************************************************************************
**				sdla_56k_alaram_print()	
**
** Description: 
** Arguments:
** Returns:
*/
static void sdla_56k_alarm_print(void* card_id, unsigned long status)
{
	sdla_t*		card = (sdla_t*)card_id;
	unsigned short	alarms = (unsigned short)status;

	if (!alarms){
		alarms = sdla_56k_alarm(card_id, 0);
	}
	
	DEBUG_EVENT("%s: 56K Alarms:\n", card->devname);
 	DEBUG_EVENT("%s: In Service:            %s Data mode idle:    %s\n",
				card->devname,
			 	INS_ALARM_56K(alarms), 
			 	DMI_ALARM_56K(alarms));
	
 	DEBUG_EVENT("%s: Zero supp. code:       %s Ctrl mode idle:    %s\n",
				card->devname,
			 	ZCS_ALARM_56K(alarms), 
			 	CMI_ALARM_56K(alarms));

 	DEBUG_EVENT("%s: Out of service code:   %s Out of frame code: %s\n",
				card->devname,
			 	OOS_ALARM_56K(alarms), 
			 	OOF_ALARM_56K(alarms));
		
 	DEBUG_EVENT("%s: Valid DSU NL loopback: %s Unsigned mux code: %s\n",
				card->devname,
			 	DLP_ALARM_56K(alarms), 
			 	UMC_ALARM_56K(alarms));

 	DEBUG_EVENT("%s: Rx loss of signal:     %s \n",
				card->devname,
			 	RLOS_ALARM_56K(alarms)); 
	return;
}
