/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_ppp.c,v 1.13 2004/07/09 16:36:11 sangoma Exp $
 */

/*****************************************************************************
 * sdla_ppp.c	WANPIPE(tm) Multiprotocol WAN Link Driver. PPP module.
 *
 * Author: 	Alex Feldman <al.feldman@@sangoma.com>
 *
 * ============================================================================
 *
 * Sep 17, 2003	 Alex Feldman		o Fixed ALTQ bug (IFQ_ENQUEUE)
 * Mar 10, 2000  Alex Feldman RX_CORR	o If receive buffer corrupted in 
 * 					  rx_intr(), disable interrupt!!!
 * Nov 21, 1999  Alex Feldman		o Updated for FreeBSD 3.2
 * Jan 06, 1997	 Gene Kozin		o Initial version.
 *****************************************************************************
 */
/*
****************************************************************************
****		For Debug purpose (only OpenBSD)			****
****************************************************************************
*/
#ifdef __OpenBSD__
# undef WANDEBUG_PPP	/* Uncomment this line for Debug purpose */
# undef ENABLE_IPV6	/* Uncomment this line for Enabling IPv6 support */
#endif

#include <net/wanpipe_config.h>
#include <net/wanpipe_includes.h>
#define _GNUC_
#include <net/wanpipe_snmp.h>
#include <net/wanpipe.h>		/* WANPIPE common user API definision */
#include <net/if_wanpipe_common.h>
#include <net/sdla_ppp.h>		/* PPP firmware API definitions */
/****** Defines & Macros ****************************************************/

#ifdef	_DEBUG_
#define	STATIC
#define WANDEBUG_PPP
#else
#define	STATIC		static
#endif

#define	PPP_DFLT_MTU	1500	/* default MTU */
#define	PPP_MAX_MTU	4000	/* maximum MTU */
#define PPP_HDR_LEN	1

#define HZ          hz

#define	CONNECT_TIMEOUT	(90*HZ)	/* link connection timeout */
#define	HOLD_DOWN_TIME	(5*HZ)	/* link hold down time : Changed from 30 to 5 */

/* For handle_IPXWAN() */
#define CVHexToAscii(b) (((unsigned char)(b) > (unsigned char)9) ? ((unsigned char)'A' + ((unsigned char)(b) - (unsigned char)10)) : ((unsigned char)'0' + (unsigned char)(b)))

/* Macro for enabling/disabling debugging comments */

#define DCD(a)   ( a & WAN_DCD_HIGH ? "HIGH" : "LOW" )
#define CTS(a)   ( a & WAN_CTS_HIGH ? "HIGH" : "LOW" ) 
#ifdef ENABLE_IPV6
# define LCP(a)  ( (a == LCP_STATE_OPENED) ? "OPEN" : "CLOSED" )
# define IP(a)   ( (a == IP_STATE_OPENED) ? "ENABLED" : "DISABLED" )
#else
# define LCP(a)  ( a == 0x09 ? "OPEN" : "CLOSED" )
# define IP(a)   ( a == 0x09 ? "ENABLED" : "DISABLED" )
#endif

#define TMR_INT_ENABLED_UPDATE  	0x0001
#define TMR_INT_ENABLED_PPP_EVENT	0x0002
#define TMR_INT_ENABLED_UDP		0x0004
#define TMR_INT_ENABLED_TE		0x0008
#define TMR_INT_ENABLED_CONFIG		0x0010
#define TMR_INT_ENABLED_TEST		0x0020
#define TMR_INT_ENABLED_PROCESS_ROUTE	0x0040

/* Set Configuraton Command Definitions */
#define PERCENT_TX_BUFF			60
#define TIME_BETWEEN_CONF_REQ  		30
#define TIME_BETWEEN_PAP_CHAP_REQ	30
#define WAIT_PAP_CHAP_WITHOUT_REPLY     300
#define WAIT_AFTER_DCD_CTS_LOW          5
#define TIME_DCD_CTS_LOW_AFTER_LNK_DOWN 10
#define WAIT_DCD_HIGH_AFTER_ENABLE_COMM 900
#define MAX_CONF_REQ_WITHOUT_REPLY      10
#define MAX_TERM_REQ_WITHOUT_REPLY      2
#define NUM_CONF_NAK_WITHOUT_REPLY      5
#define NUM_AUTH_REQ_WITHOUT_REPLY      10

#define END_OFFSET 0x1F0

#ifdef ENABLE_IPV6
# define WAN_IPV4	4
# define WAN_IPV6	6
# define ip6_local	local_addr.addr6
# define ip6_remote	remote_addr.addr6
# define IPVER2PROTOCOL(ver)	(ver == 0x04) ? htons(ETHERTYPE_IP) : htons(ETHERTYPE_IPV6)
#endif
/******Data Structures*****************************************************/

/* This structure is placed in the private data area of the device structure.
 * The card structure used to occupy the private area but now the following 
 * structure will incorporate the card structure along with PPP specific data
 */
  
typedef struct ppp_private_area
{
	wanpipe_common_t	common;
	unsigned long		router_start_time;	/*router start time in sec */
	unsigned long		tick_counter;		/*used for 5 second counter*/
	unsigned mc;					/*multicast support on or off*/
#ifdef ENABLE_IPV6
	unsigned char		enable_IPv4;
	unsigned char		enable_IPv6;
#endif
	unsigned char		enable_IPX;
	unsigned long		network_number;
	unsigned char		pap;
	unsigned char		chap;
	unsigned char		sysname[31];		/* system name for in-bnd auth*/
#ifdef ENABLE_IPV6
	unsigned char		chap_userid[511];	/* list of user ids */
	unsigned char		chap_passwd[511];	/* list of passwords */
	unsigned char		pap_userid[511];	/* list of user ids */
	unsigned char		pap_passwd[511];	/* list of passwords */
#else
	unsigned char		userid[511];		/* list of user ids */
	unsigned char		passwd[511];		/* list of passwords */
#endif
	unsigned		protocol;		/* Mbuff Protocol */
	uint32_t		ip_local;		/* Local IP Address */
	uint32_t		ip_remote;		/* remote IP Address */
#ifdef ENABLE_IPV6
	union {
		struct in_addr  addr;
		struct in6_addr addr6;
	} local_addr;
	union {
		struct in_addr  addr;
		struct in6_addr addr6;
	} remote_addr;
#endif

	unsigned char		timer_int_enabled;	/* Who enabled the timer inter*/
	unsigned char		update_comms_stats;	/* Used by update function */
	unsigned long		curr_trace_addr;	/* Trace information */
	unsigned long		start_trace_addr;
	unsigned long		end_trace_addr;

	unsigned short		udp_pkt_lgth;
	char			udp_pkt_src;
      	char			udp_pkt_data[sizeof(wan_udp_pkt_t)];

	/* PPP specific statistics */
	if_send_stat_t		if_send_stat;
	rx_intr_stat_t		rx_intr_stat;
	pipe_mgmt_stat_t	pipe_mgmt_stat;

	unsigned long		router_up_time; 

#ifdef ENABLE_IPV6
	unsigned char		initialized;
#endif
	/* WANPIPE debugging statistics */
	unsigned long		rx_lcp;
	unsigned long		rx_conf_nak;

} ppp_private_area_t;

/* variable for keeping track of enabling/disabling FT1 monitor status */
static int rCount = 0;
#ifdef __OpenBSD__
extern int ticks;
#endif
/****** Function Prototypes *************************************************/

/* WAN link driver entry points. These are called by the WAN router module. */
static int update(wan_device_t *wandev);
static int new_if(wan_device_t *wandev, struct ifnet *ifp, wanif_conf_t *conf);
static int del_if(wan_device_t *wandev, struct ifnet *ifp);

/* WANPIPE-specific entry points */
static int wpp_exec (struct sdla *card, void *u_cmd, void *u_data);

/* Network device interface */
static int ppp_close  (struct ifnet *ifp);
/* ifnet entry points */
static 	int 	ppp_output		/* output routine (enqueue) */
		__P((struct ifnet *, struct mbuf *, struct sockaddr *,
		     struct rtentry *));
static 	void	ppp_start		/* initiate output routine */
		__P((struct ifnet *));
static	int	ppp_ioctl		/* ioctl routine */
		__P((struct ifnet *, u_long, caddr_t));
static	void	ppp_watchdog		/* timer routine */
		__P((struct ifnet *));
static	int	ppp_init		/* init routine */
		__P((struct ifnet *, struct ifaddr *));

/* PPP firmware interface functions */
static int ppp_read_version(sdla_t *card, char *str);
#ifdef ENABLE_IPV6
static int ppp_set_outbnd_chap_auth(sdla_t *card, ppp_private_area_t *ppp_priv_area);
static int ppp_set_inbnd_chap_auth(sdla_t *card, ppp_private_area_t *ppp_priv_area);
static int ppp_set_outbnd_pap_auth(sdla_t *card, ppp_private_area_t *ppp_priv_area);
static int ppp_set_inbnd_pap_auth(sdla_t *card, ppp_private_area_t *ppp_priv_area);
#else
static int ppp_set_outbnd_auth(sdla_t *card, ppp_private_area_t *ppp_priv_area);
static int ppp_set_inbnd_auth(sdla_t *card, ppp_private_area_t *ppp_priv_area);
#endif
static int ppp_configure(sdla_t *card, void *data);
static int ppp_set_intr_mode(sdla_t *card, unsigned char mode);
#if 0
static int ppp_read_intr_mode(sdla_t *card);
#endif
static int ppp_comm_enable(sdla_t *card);
static int ppp_comm_disable(sdla_t *card);
static int ppp_get_err_stats(sdla_t *card, ppp_err_stats_t* stats);
static int ppp_get_stats(sdla_t *card, ppp508_stats_t* stats);
static int ppp_send(sdla_t *card, void *data, unsigned len, unsigned proto);
static int ppp_error(sdla_t *card, int err, wan_mbox_t*);
static int set_adapter_config (sdla_t* card);

static void wpp_isr(sdla_t* card);
static void rx_intr(sdla_t *card);
static void event_intr(sdla_t *card);
static void timer_intr(sdla_t *card);

/* Background polling routines */
static void process_route(netdevice_t*);
static void poll_disconnected(sdla_t *card);

/* Miscellaneous functions */
static int read_info( sdla_t *card );
static int read_connection_info (sdla_t *card);
static int remove_route( sdla_t *card );
static int config508(ppp_private_area_t *ppp_priv_area, sdla_t *card);
#ifdef ENABLE_IPV6
static int ppp_get_ip_cfg(ppp_private_area_t*, ppp508_conf_t*, struct ifaddr*);
#endif
static void show_disc_cause(sdla_t * card, unsigned cause);
static int process_udp_mgmt_pkt(sdla_t *card, struct ifnet *ifp, 
				ppp_private_area_t *ppp_priv_area);
static void init_ppp_tx_rx_buff( sdla_t *card );
#ifdef PPP_OLD_INTR_TEST
static int ppp_intr_test( sdla_t *card );
#endif
static int udp_pkt_type(sdla_t* card, struct mbuf *m0);
static void init_ppp_priv_struct( ppp_private_area_t *ppp_priv_area);
static void init_global_statistics( sdla_t *card );
static int tokenize(char *str, char **tokens);
#if 0
static char* strstrip(char *str, char *s);
#endif
static int chk_bcast_mcast_addr(sdla_t* card, struct ifnet* ifp,
				struct mbuf *m);
static int Read_connection_info;
static int Intr_test_counter;
static unsigned short available_buffer_space;

/* IPX functions */
#ifdef SUPPORT_IPX
static void switch_net_numbers(unsigned char *sendpacket, 
			       unsigned long network_number, 
			       unsigned char incoming);
static int handle_IPXWAN(unsigned char *sendpacket, char *devname, 
		         unsigned char enable_PX, unsigned long network_number, 
			 unsigned short proto);
#endif

static int store_udp_mgmt_pkt(char udp_pkt_src, sdla_t* card,
                                struct mbuf *m, struct ifnet* ifp,
                                ppp_private_area_t* ppp_priv_area );

#ifdef ENABLE_IPV6
static int reply_udp_ipv6( unsigned char *data, unsigned int mbox_len);
#endif

#if 0
static char *my_strtok(char *p, const char *tok);
#endif
static char *my_strchr(char *p, char ch);

extern struct mbuf* sdla_get(struct ifnet*, char*, unsigned);

static WRITE_FRONT_END_REG_T 	write_front_end_reg;
static READ_FRONT_END_REG_T	read_front_end_reg;
static void ppp_enable_timer(void* card_id);
static void ppp_handle_front_end_state(void* card_id);

/* SNMP */
static int ppp_snmp_data(sdla_t* card, struct ifnet* ifp, void* data);

static int ppp_udp_ioctl(struct ifnet* ifp, caddr_t data);
static int process_udp_cmd(struct ifnet* ifp, wan_udp_hdr_t* udp_hdr);

static int ppp_debugging(sdla_t* card);
static unsigned long ppp_crc_frames(sdla_t* card);
static unsigned long ppp_abort_frames(sdla_t * card);
static unsigned long ppp_tx_underun_frames(sdla_t* card);
/****** Public Functions ****************************************************/

/*============================================================================
 * PPP protocol initialization routine.
 *
 * This routine is called by the main WANPIPE module during setup.  At this
 * point adapter is completely initialized and firmware is running.
 *  o read firmware version (to make sure it's alive)
 *  o configure adapter
 *  o initialize protocol-specific fields of the adapter data space.
 *
 * Return:	0	o.k.
 *		< 0	failure.
 */
int wpp_init(sdla_t *card, wandev_conf_t *conf)
{
	union{
		char str[80];
	} u;
	wan_spinlock_t	s;
	int err = 0;

	wan_spin_lock_irq(NULL, &s);
	DEBUG_INIT("%s: Starting PPP initialization\n", card->wandev.name);
	/* Verify configuration ID */
	if (conf->config_id != WANCONFIG_PPP){ 
		DEBUG_EVENT("%s: invalid configuration ID %u!\n",
				card->devname, conf->config_id);
		wan_spin_unlock_irq(NULL, &s);
		return -EINVAL;
	}

	/* Initialize miscellaneous pointers to structures on the adapter */
    	card->mbox_off = PPP514_MB_OFFS;
    	card->flags_off= PPP514_FLG_OFFS;

	card->intr_type_off = 
			card->flags_off +
			offsetof(ppp_flags_t, iflag);
	card->intr_perm_off = 
			card->flags_off +
			offsetof(ppp_flags_t, imask);

	err = card->hw_iface.check_mismatch(card->hw,conf->te_cfg.media);
	if (err){
		return err;
	}
	
	/* TE1 Make special hardware initialization for T1/E1 board */
	if (IS_TE1(conf->te_cfg)){
		
		memcpy(&card->wandev.te_cfg, &conf->te_cfg, sizeof(sdla_te_cfg_t));
		card->wandev.write_front_end_reg = write_front_end_reg;
		card->wandev.read_front_end_reg = read_front_end_reg;
		card->wandev.te_enable_timer = ppp_enable_timer;
		card->wandev.te_link_state = ppp_handle_front_end_state;
		conf->interface = 
			(card->wandev.te_cfg.media == WANOPT_MEDIA_T1) ? WANOPT_V35 : WANOPT_RS232;
		conf->clocking = WANOPT_EXTERNAL;

	}else if (IS_56K(conf->te_cfg)){

		memcpy(&card->wandev.te_cfg, &conf->te_cfg, sizeof(sdla_te_cfg_t));
		card->wandev.write_front_end_reg = write_front_end_reg;
		card->wandev.read_front_end_reg = read_front_end_reg;
		
		conf->clocking = WANOPT_EXTERNAL;
	}else{
		card->wandev.front_end_status = FE_CONNECTED;	
	}
    
	/* Read firmware version.  Note that when adapter initializes, it
	 * clears the mailbox, so it may appear that the first command was
	 * executed successfully when in fact it was merely erased. To work
	 * around this, we execute the first command twice.
	 */
	if (ppp_read_version(card, NULL) || ppp_read_version(card, u.str)){
		wan_spin_unlock_irq(NULL, &s);
		return -EIO;
	}

	DEBUG_EVENT("%s: running PPP firmware v%s\n",card->devname, u.str); 

	if (set_adapter_config(card)){
		wan_spin_unlock_irq(NULL, &s);
		return -EIO;
	}

	/* Adjust configuration and set defaults */
	card->wandev.mtu = (conf->mtu) ? min(conf->mtu, PPP_MAX_MTU) : PPP_DFLT_MTU;

	card->wandev.bps		= conf->bps;
	card->wandev.interface	= conf->interface;
	card->wandev.clocking	= conf->clocking;
	card->isr		= &wpp_isr;
	card->poll		= NULL; 
	card->exec		= &wpp_exec;
	card->wan_debugging	= &ppp_debugging;
	card->get_crc_frames	= &ppp_crc_frames;
	card->get_abort_frames	= &ppp_abort_frames;
	card->get_tx_underun_frames = &ppp_tx_underun_frames;
	card->wandev.update	= &update;
	card->wandev.new_if	= &new_if;
	card->wandev.del_if	= &del_if;
	card->wandev.udp_port  	= conf->udp_port;
	card->wandev.ttl	= conf->ttl;
	card->irq_dis_if_send_count = 0;
	card->irq_dis_poll_count 	= 0;
	card->u.p.authenticator 	= conf->u.ppp.authenticator;
	card->u.p.ip_mode 		= 
			conf->u.ppp.ip_mode ? conf->u.ppp.ip_mode : WANOPT_PPP_STATIC;
	card->TracingEnabled    	= 0;
	Read_connection_info    	= 1;

	/* initialize global statistics */
	init_global_statistics(card);

	/* Update card state */
	card->wandev.state	= WAN_DISCONNECTED;
	/* SNMP data */
	card->get_snmp_data = &ppp_snmp_data;

	wan_spin_unlock_irq(NULL, &s);
#ifdef PPP_OLD_INTR_TEST
	return ppp_intr_test(card);
#endif
	return 0;
}

/******* WAN Device Driver Entry Points *************************************/

/*============================================================================
 * Update device status & statistics.
 */
static int update(wan_device_t *wandev)
{
	netdevice_t		*ifp;
	sdla_t 			*card = wandev->private;
	volatile ppp_private_area_t *ppp_priv_area = NULL;
	unsigned long 		timeout;

	/* sanity checks */
	if ((wandev == NULL) || (wandev->private == NULL)){
		return -EFAULT;
	}

	if (wandev->state == WAN_UNCONFIGURED){
		return -ENODEV;
	}

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	if (ifp == NULL)
		return -ENODEV;
	if((ppp_priv_area = wan_netif_priv(ifp)) == NULL){
		return -ENODEV;
	}
	ppp_priv_area->update_comms_stats = 2;
	ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_UPDATE;
	card->hw_iface.set_bit(card->hw, card->intr_perm_off, PPP_INTR_TIMER);

	/* wait a maximum of 1 second for the statistics to be updated */ 
	timeout = SYSTEM_TICKS;
	for(;;){ 
		if (ppp_priv_area->update_comms_stats == 0){
			break;
		}
		if ((SYSTEM_TICKS - timeout) > (1 * HZ)){
			ppp_priv_area->update_comms_stats = 0;
			ppp_priv_area->timer_int_enabled &=
					~TMR_INT_ENABLED_UPDATE; 
			return -EAGAIN;
		}
	}

	return 0;
}

/*============================================================================
 * Create new logical channel.
 * This routine is called by the router when ROUTER_IFNEW IOCTL is being
 * handled.
 * o parse media- and hardware-specific configuration
 * o make sure that a new channel can be created
 * o allocate resources, if necessary
 * o prepare network device structure for registaration.
 *
 * Return:	0	o.k.
 *		< 0	failure (channel will not be created)
 */
static int new_if(wan_device_t *wandev, struct ifnet *ifp, wanif_conf_t *conf)
{
	sdla_t*		card = wandev->private;
	ppp_private_area_t* ppp_priv_area;
	
	if (wandev->ndev)
		return -EEXIST;

	if ((conf->name[0] == '\0') || (strlen(conf->name) > WAN_IFNAME_SZ)){ 
		DEBUG_INIT("%s: invalid interface name!\n", card->devname);
		return -EINVAL;
	}

	/* allocate and initialize private data */
	ppp_priv_area = wan_malloc(sizeof(ppp_private_area_t));

	if (ppp_priv_area == NULL)
		return	-ENOMEM;

	bzero(ppp_priv_area, sizeof(ppp_private_area_t));

	ppp_priv_area->common.card = card; 

	if (wandev->new_if_cnt > 0){
		wan_free(ppp_priv_area);
		return -EEXIST;
	}

	/* initialize data */
	memcpy(card->u.p.if_name, conf->name, strlen(conf->name)); 

	/* increment number of interfaces */
	wandev->new_if_cnt++;

	/* initialize data in ppp_private_area structure */
	init_ppp_priv_struct(ppp_priv_area);

	ppp_priv_area->mc = conf->mc;
	ppp_priv_area->pap = conf->pap;
	ppp_priv_area->chap = conf->chap;

#ifdef ENABLE_IPV6

	/* If no userid or no passwords are specified for chap */
	if (ppp_priv_area->chap){
   		if (!(strlen(conf->chap_userid) && strlen(conf->chap_passwd))){
			wan_free(ppp_priv_area);
			return -EINVAL;
    		}
	}

	/* If no userid or no passwords are specified for pap */
	if (ppp_priv_area->pap){
   		if (!(strlen(conf->pap_userid) && strlen(conf->pap_passwd))){
			wan_free(ppp_priv_area);
			return -EINVAL;
    		}
	}

#else

	/* If no user ids are specified */
	if (!strlen(conf->userid) && (ppp_priv_area->pap||ppp_priv_area->chap)){
		wan_free(ppp_priv_area);
		return -EINVAL;
	}

	/* If no passwords are specified */
	if (!strlen(conf->passwd) && (ppp_priv_area->pap||ppp_priv_area->chap)){
		wan_free(ppp_priv_area);
		return -EINVAL;
	}
#endif

	if (strlen(conf->sysname) > 31){
		wan_free(ppp_priv_area);
		return -EINVAL;
	}

	/* If no system name is specified */
	if (!strlen(conf->sysname) && (card->u.p.authenticator)){
		wan_free(ppp_priv_area);
		return -EINVAL;
	}

	/* copy the data into the ppp private structure */
#ifdef ENABLE_IPV6
	bcopy(conf->chap_userid, 
		ppp_priv_area->chap_userid, 
		strlen(conf->chap_userid));
	bcopy(conf->chap_passwd, 
		ppp_priv_area->chap_passwd, 
		strlen(conf->chap_passwd));
	bcopy(conf->pap_userid, 
		ppp_priv_area->pap_userid, 
		strlen(conf->pap_userid));
	bcopy(conf->pap_passwd, 
		ppp_priv_area->pap_passwd, 
		strlen(conf->pap_passwd));
#else
	bcopy(conf->userid, ppp_priv_area->userid, strlen(conf->userid));
	bcopy(conf->passwd, ppp_priv_area->passwd, strlen(conf->passwd));
#endif
	bcopy(conf->sysname, ppp_priv_area->sysname, strlen(conf->sysname));

#ifdef ENABLE_IPV6
	ppp_priv_area->enable_IPv4 = conf->enable_IPv4;
	ppp_priv_area->enable_IPv6 = conf->enable_IPv6;
#endif
	ppp_priv_area->enable_IPX = conf->enable_IPX;
	if (conf->network_number){
		ppp_priv_area->network_number = conf->network_number;
	}else{
		ppp_priv_area->network_number = 0xDEADBEEF;
	}

	/* prepare network device data space for registration */
	if (wan_netif_init(ifp, card->u.p.if_name)){
		DEBUG_EVENT("%s:%s: Failed to initialize network interface!\n", 
					card->devname,
					card->u.p.if_name);
		wan_free(ppp_priv_area);
		return -ENOMEM;
	}
	ifp->if_softc          	= ppp_priv_area;
	ifp->if_mtu            	= wandev->mtu; /* ETHERMTU */
	ifp->if_type           	= IFT_OTHER;
	ifp->if_flags          	= IFF_POINTOPOINT; 
	if (ppp_priv_area->mc){
		ifp->if_flags |= IFF_MULTICAST;
	}
#ifdef ENABLE_IPV6
	ifp->if_flags          |= IFF_MULTICAST; 
#endif
	ifp->if_output		= ppp_output;
	ifp->if_start          	= ppp_start;
	ifp->if_watchdog       	= ppp_watchdog;
#ifdef __FreeBSD__
	ifp->if_init		= (void*)ppp_init;
#endif
	ifp->if_ioctl		= ppp_ioctl;
	ifp->if_snd.ifq_len	= 0;
	
	/* ???? This members was not initialize in 3.1 */
	ifp->if_hdrlen = PPP_HDR_LEN;

	return 0;
}

/* 
 * ============================================================================
 * Delete logical channel.
 */
static int del_if (wan_device_t* wandev, struct ifnet* ifp)
{
	sdla_t*	card = wandev->private;

	DEBUG_INIT("%s: Deleting PPP interface %s...\n", 
			card->devname, card->u.p.if_name); 
	if (card == NULL)
		return -ENODEV;

	/* Initialize interrupt handler pointer */
	card->isr = NULL;

	/* Close current interface, set interrupt permission byte to 0 */ 
	ppp_close(ifp);

	/* Initialize network interface */
	IFQ_PURGE(&ifp->if_snd);
	ifp->if_output = NULL;
	ifp->if_start = NULL;
	ifp->if_ioctl = NULL;
	ifp->if_watchdog = NULL;
	ifp->if_flags &= ~(IFF_RUNNING | IFF_UP | IFF_OACTIVE);
	if (ifp->if_softc){
        	wan_free(ifp->if_softc);
		ifp->if_softc = NULL;
	}
	wan_netif_del(ifp);

	return 0;
}

/****** WANPIPE-specific entry points ***************************************/

/*============================================================================
 * Execute adapter interface command.
 */

/* FIXME: Why do we need this ???? */
static int wpp_exec(struct sdla *card, void *u_cmd, void *u_data)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 		len, err;

	if (!u_cmd || copyin(u_cmd, mbox, sizeof(wan_cmd_t))){
		return -EFAULT;
	}
	len = mbox->wan_data_len;

	if (len){
		if (copyin(u_data, mbox->wan_data, len)){ 
			return -EFAULT;
		}
	}

	/* execute command */
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	/* return result */
	if (copyout(&mbox->wan_command, u_cmd, sizeof(wan_cmd_t)))
		return -EFAULT;   
	len = mbox->wan_data_len;

	if (len && u_data){
		if (copyout(mbox->wan_data, u_data, len)){
			return -EFAULT;
		}
	}
	return 0;
}


/****** Network Device Interface ********************************************/

/*
 * ptpipe_ioctl() -
 *
 */
static int
ppp_ioctl(struct ifnet * ifp, u_long command, caddr_t data)
{
	ppp_private_area_t *ppp_priv_area = ifp->if_softc; 
	sdla_t             *card = ppp_priv_area->common.card;
	struct ifaddr      *ifa = (struct ifaddr *)data; /* address struct */
	wan_spinlock_t		s;	    	    /* saved interrupt state */
	int		       error = 0;    	    /* resulting status */
#ifdef ENABLE_IPV6
	struct ifreq* 	ifr = NULL;
	struct sockaddr_in* sin = NULL;
#endif

	wan_spin_lock_irq(NULL, &s);

	switch (command){ 
	case SIOCSIFADDR:
		DEBUG_IOCTL("%s: Ioctl (SIOCSIFADDR) called.\n", 
					card->devname);
		ifp->if_flags |= IFF_UP;
		/* check the protocol family - we currently support IP only */
		switch(ifa->ifa_addr->sa_family){ 
		case AF_INET:
#ifdef ENABLE_IPV6
			if (!ppp_priv_area->initialized)
#endif
				ppp_init(ifp, ifa);
			break;

#ifdef ENABLE_IPV6
		case AF_INET6:
			if (!ppp_priv_area->initialized){
				ppp_init(ifp, ifa);
			}else{
				/* Copy IPv6 addresses to private structure */
				ppp_get_ip_cfg(ppp_priv_area, NULL, ifa);
			}
			break;
#endif

		default:
			error = EAFNOSUPPORT;
			break;
		}
		break;

	case SIOCGIFADDR:
		DEBUG_IOCTL("%s: Ioctl (SIOCGIFADDR) called.\n", 
					card->devname);
		break;

	case SIOCSIFFLAGS:
		DEBUG_IOCTL("%s: Ioctl (SIOCSIFFLAGS) called.\n", 
					card->devname);
		/*
		 * If the interface is marked up - enable communications. If down - 
		 * disable communications.  IFF_UP is taken care of before entering 
		 * this function.
		 */
		if (((ifp->if_flags & IFF_UP)==0)/*&& (ifp->if_flags & IFF_RUNNING)*/){ 
			/* bring it down */
			DEBUG_EVENT("%s: Bringing interface down.\n", 
					card->devname);
			ifp->if_flags &= ~IFF_RUNNING;
			/* set up the disconnect cause */
		}else{ /* bring it up */ 
	
			DEBUG_EVENT("%s: Bringing interface up.\n", 
					card->devname);
			/* update the up time ... BF-0003 */
  	    		wan_getcurrenttime(&ppp_priv_area->router_up_time, NULL);
			/* Start debugging sequence */
			WAN_DEBUG_START(card);
		}
		break;

	case SIOCGIFFLAGS:
    		DEBUG_IOCTL("%s: Ioctl (SIOCGIFFLAGS) called.\n", 
				card->devname);
		break;

	case SIOCSIFMTU:
		DEBUG_IOCTL("%s: Ioctl (SIOCSIFMTU) called.\n", 
				card->devname);
		error = EINVAL;
		break;

#ifdef ENABLE_IPV6
	case SIOCGIFDSTADDR:
		DEBUG_IOCTL("%s: Ioctl (SIOCGIFDSTADDR) called.\n", 
				card->devname);
		ifr = (struct ifreq*)data;
		sin = (struct sockaddr_in*)&ifr->ifr_ifru.ifru_dstaddr;
		if (sin->sin_family == AF_INET){
			sin->sin_family = AF_INET;
			sin->sin_addr = ppp_priv_area->remote_addr.addr;
		}else{
			struct sockaddr_in6 sin6; 
			sin6.sin6_family = AF_INET6;
			sin6.sin6_scope_id = 0xc;
    			sin6.sin6_addr = ppp_priv_area->ip6_remote;
			if (copyout((void*)&sin6, (void*)ifr->ifr_data, sizeof(struct sockaddr_in6))){
				DEBUG_EVENT("%s: Failed to copy IP address to user space\n",
							card->devname);
				error = EINVAL;
				break;
			}
		}
		break;
#endif

	case SIOC_WANPIPE_PIPEMON:
    		DEBUG_IOCTL("%s: Ioctl (SIOC_WANPIPE_PIPEMON) called.\n", 
				card->devname);
		error = ppp_udp_ioctl(ifp, data);
		break;

	case SIOC_WANPIPE_SNMP:
	case SIOC_WANPIPE_SNMP_IFSPEED:
		wan_spin_unlock_irq(NULL, &s);
		return wan_snmp_data(card, ifp, command, (struct ifreq*)data);
		break;

	default:
		DEBUG_IOCTL("%s: Ioctl (default) called.\n", 
				card->devname);
		error = EINVAL;
		break;
	}

	wan_spin_unlock_irq(NULL, &s);
	return error;
}

/*
 * ppp_watchdog() - this function will be called for the first time
 *		    a second after first protocol-open interrupt. 
 *		    After that it will reschedule itself to be called
 *		    once every second. 
 *		    The routine will check the protocol state and 
 *		    update the interface state if it changes. Actually
 *		    the only change we intereseted in is when the protocol
 *		    goes down ( coz when it goes up we will receive the 
 *		    protocol-open interrupt ).
 */
static void ppp_watchdog(struct ifnet * ifp)
{
	ppp_private_area_t *ppp_priv_area = ifp->if_softc;
	sdla_t *card = ppp_priv_area->common.card;
    
	if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_PROCESS_ROUTE){
		process_route(ifp);
		ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_PROCESS_ROUTE;
	}
	
#ifndef PPP_OLD_INTR_TEST
	if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_TEST){
		if (Intr_test_counter < MAX_NEW_INTR_TEST_COUNTER){
			ppp_priv_area->timer_int_enabled &= 
						~TMR_INT_ENABLED_TEST;
			DEBUG_EVENT("%s: Interrupt Test Failed, Counter: %d\n",
					card->devname, Intr_test_counter);
			DEBUG_EVENT("%s: Please choose another interrupt!!\n", 
					card->devname); 
			ppp_set_intr_mode(card, 0x00);
		}
	}
#endif
	
	return;
}


/* 
 * ============================================================================
 * Open network interface.
 * o enable communications and interrupts.
 * o prevent module from unloading by incrementing use count
 *
 * Return 0 if O.k. or errno.
 */
static int ppp_init (struct ifnet *ifp, struct ifaddr *ifa)
{
	ppp_private_area_t*	ppp_priv_area = ifp->if_softc;
	sdla_t*			card = ppp_priv_area->common.card;
	struct sockaddr_in*	saddr = NULL;
	int			err = 0;

	DEBUG_INIT("%s: Initialize interface...\n", 
				card->devname);

	switch (card->u.p.ip_mode){ 
	case WANOPT_PPP_STATIC:
	case WANOPT_PPP_HOST:
		/* Local IP address */
		saddr = (struct sockaddr_in*)ifa->ifa_addr;
		if (saddr){
			ppp_priv_area->ip_local = saddr->sin_addr.s_addr;
		}
		/* Destination IP address */
		saddr = (struct sockaddr_in*)ifa->ifa_dstaddr;
		if (saddr){
			ppp_priv_area->ip_remote = saddr->sin_addr.s_addr;
		}
		break;

	case WANOPT_PPP_PEER:
		ppp_priv_area->ip_local		= 0x00;
		ppp_priv_area->ip_remote	= 0x00;
		break;
	}

#ifdef PPP_OLD_INTR_TEST
	ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_CONFIG;
#else
	Intr_test_counter = 0;
	ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_TEST;
	/* Enable watchdog timer in case if hardware interrupt will failed */ 
	ifp->if_timer = 5;
#endif
	/* Turn on timer intrrupt */
	if (ppp_set_intr_mode(card, PPP_INTR_TIMER)){
		return -EIO;
	}


	wanpipe_open(card);
	return err;
}

/*============================================================================
 * Close network interface.
 * o if this is the last open, then disable communications and interrupts.
 * o reset flags.
 */
static int ppp_close(struct ifnet *ifp)
{
	ppp_private_area_t*	ppp_priv_area = ifp->if_softc;
	sdla_t*			card = ppp_priv_area->common.card;
	wan_spinlock_t		s;

	wan_spin_lock_irq(NULL, &s);
	
	wanpipe_close(card);
	wanpipe_set_state(card, WAN_DISCONNECTED);
	ppp_priv_area->common.state = WAN_DISCONNECTED;
	ppp_set_intr_mode(card, 0);
	ppp_comm_disable(card);

	if (IS_TE1_CARD(card)){
    		sdla_te_unconfig(card);
	}
#ifdef ENABLE_IPV6
	ppp_priv_area->initialized = 0;
#endif

	wan_spin_unlock_irq(NULL, &s);
	return 0;
}

/* 
 * ============================================================================
 * Configure PPP interface.
 * o enable communications and interrupts.
 * o prevent module from unloading by incrementing use count
 *
 * Return 0 if O.k. or errno.
 */
static int ppp_config (struct ifnet* ifp)
{
	ppp_private_area_t*	ppp_priv_area = ifp->if_softc;
	sdla_t*			card = ppp_priv_area->common.card;
	int			err = 0;

	DEBUG_INIT("%s: Initialize interface...\n", 
				card->devname);

	if (IS_TE1_CARD(card)) {
		DEBUG_EVENT("%s: Configuring onboard %s CSU/DSU\n",
					card->devname, 
					(IS_T1_CARD(card))?"T1":"E1");
		if (sdla_te_config(card)){
			DEBUG_EVENT("%s: Failed %s configuratoin!\n",
				card->devname,
				(IS_T1_CARD(card))?"T1":"E1");
			return 0;
		}
		
	}else if (IS_56K_CARD(card)){
		DEBUG_EVENT("%s: Configuring 56K onboard CSU/DSU\n",
			card->devname);

		if(sdla_56k_config(card)) {
			DEBUG_EVENT("%s: Failed 56K configuration!\n",
				card->devname);
			return 0;
		}
    	}
	if (config508(ppp_priv_area, card)){
		return -EIO;
	}

	/* Initialize Rx/Tx buffer control fields */
	init_ppp_tx_rx_buff( card );
	
	if (ppp_set_intr_mode(card, PPP_INTR_RXRDY|
    				PPP_INTR_TXRDY|
	    			PPP_INTR_MODEM|
	    			PPP_INTR_DISC |
	    			PPP_INTR_OPEN |
	    			PPP_INTR_DROP_DTR |
				PPP_INTR_TIMER)){ 
		return -EIO;
	}

	/* Turn off the transmit and timer interrupt */
	card->hw_iface.clear_bit(card->hw, card->intr_perm_off, (PPP_INTR_TXRDY | PPP_INTR_TIMER)); 

	/* If you are not the authenticator and any one of the protocol is 
	 * enabled then we call the set_out_bound_authentication.
	 */
#ifdef ENABLE_IPV6

	if (!card->u.p.authenticator){
		if (ppp_priv_area->pap && ppp_set_outbnd_pap_auth(card, ppp_priv_area)){
			return -EIO;
		}
		if (ppp_priv_area->chap && ppp_set_outbnd_chap_auth(card, ppp_priv_area)){
			return -EIO;
		}	
	}else{ 
		if (ppp_priv_area->pap && ppp_set_inbnd_pap_auth(card, ppp_priv_area)){
			return -EIO;
		}
		if (ppp_priv_area->chap && ppp_set_inbnd_chap_auth(card, ppp_priv_area)){
			return -EIO;
		}
	} 

#else

	if (!card->u.p.authenticator  && (ppp_priv_area->pap || ppp_priv_area->chap)){
		if (ppp_set_outbnd_auth(card, ppp_priv_area)){
			return -EIO;
		}
	} 

	/* If you are the authenticator and any one of the protocol is enabled
	 * then we call the set_in_bound_authentication.
	 */
	if (card->u.p.authenticator && (ppp_priv_area->pap || ppp_priv_area->chap)){
		if (ppp_set_inbnd_auth(card, ppp_priv_area)){
			return -EIO;
		}
	}

	if (ppp_comm_enable(card)){
		return -EIO;
	}

#endif

	wanpipe_set_state(card, WAN_CONNECTING);
	ppp_priv_area->common.state = WAN_CONNECTING;
	wan_getcurrenttime(&ppp_priv_area->router_start_time, NULL);
#ifdef ENABLE_IPV6
	ppp_priv_area->initialized = 1;
#endif
	return err;
}

/*============================================================================
 * ptpipe_output() - output routine. The function tries to enqueue the 
 *		     packet and then if the device is not IFF_OATIVE
 *		     calls ptpipe_start(). The function returns an error
 *		     value or 0 on success.
 */
static int
ppp_output(struct ifnet *ifp, struct mbuf *m0, struct sockaddr *dst, struct rtentry *rt0)
{
	ppp_private_area_t*	ppp_priv_area = ifp->if_softc;
	sdla_t*		card = ppp_priv_area->common.card;
	int 		udp_type = 0;
	int 		error = 0;
	struct ip_packet*	ip_pckt;	/* used to interpret the data*/
#if defined(ALTQ)
	struct altq_pktattr	pktattr;
#endif

	/* we used to check here for device being UP and RUNNING - but since  */
	/* we would like to have a monitor running even if the device is down */
	/* we have to move this check till after the UDP check 		*/

	++ppp_priv_area->if_send_stat.if_send_entry;

	if (m0 == NULL){ 
		/* If we get here, some higher layer thinks we've missed an
		 * tx-done interrupt.
		 */
		DEBUG_EVENT("%s: interface %s got kicked!\n", 
				card->devname, card->u.p.if_name);
		++ppp_priv_area->if_send_stat.if_send_skb_null;
		return 0;
	}

	if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)){
		wan_skb_free(m0);	/*m_freem(m0);*/
		return ENETDOWN;
	}

	/* check the protocol family - we currently support only IP */
	if (dst->sa_family != AF_INET && dst->sa_family != AF_INET6){
        	/* protocol family is not supported */
	        DEBUG_EVENT("%s: Protocol family (%d) is not supported!\n",
				card->devname, dst->sa_family);
		wan_skb_free(m0);	/*m_freem(m0);*/
        	return EAFNOSUPPORT;
	}	

	if ((ifp->if_flags & IFF_OACTIVE) != 0){ 
		/* If our device stays busy for at least 5 seconds then we will
		 * increment statistics. We expect that device never stays busy more than
		 * 5 seconds.
		 */
		if ((SYSTEM_TICKS - ppp_priv_area->tick_counter) > (5 * HZ)){ 
			ifp->if_collisions ++;
			ppp_priv_area->if_send_stat.if_send_tbusy++;
			DEBUG_EVENT("%s: Transmit times out!\n", card->devname);
		}
	}

	/* 
	 * Ok, UDP management additions - we need to check a few things in order 
	 * to be sure that this is our UDP management packet :
	 *
	 *  o this is AF_INET packet
	 *  o this is UDP packet ( IPPCKT_IP_PROTO byte set to IPPROTO_UDP )
	 *  o the UDP port corresponds to the one set for the driver
	 *  o the UDP signature matches ours 
	 *
	 * To improve performance we will start checking with less-likely possible
	 * event ( i.e. bottom-up in the above list ). Also since mbuf is at least
	 * 128 bytes length - i don't even check for the possibility of a very short
	 * packet ( like raw IP with no data at all ) - even if this happens we wont
	 * get an exception since there's at least 128 bytes of allocated memery
	 *
	 * Oops, also i need to make sure its a request packet !!!
	 *
	 */
	ip_pckt = mtod(m0, struct ip_packet *);

	udp_type = udp_pkt_type(card, m0);
	if (udp_type == UDP_PTPIPE_TYPE){
		wan_bpf_report(ifp, m0, 1);
		if (store_udp_mgmt_pkt(UDP_PKT_FRM_STACK, card, m0, ifp, ppp_priv_area)){
			card->hw_iface.set_bit(card->hw, card->intr_perm_off, PPP_INTR_TIMER);
		}
		++ppp_priv_area->if_send_stat.if_send_PIPE_request;
		return 0;
	}

	/* Check for broadcast and multicast addresses 
	 * If found, drop (deallocate) a packet and return.
	 */
#ifndef ENABLE_IPV6
	if (chk_bcast_mcast_addr(card, ifp, m0)){
	       	IF_DROP(&ifp->if_snd);
		wan_skb_free(m0);	/*m_freem(m0);*/
		return 0;
	}
#endif

#ifdef ENABLE_IPV6
	/* check if ip version is enabled for PPP */
	if (!((dst->sa_family == AF_INET && flags->ip_state == IP_STATE_OPENED) || 
              (dst->sa_family == AF_INET6 && flags->ipv6_state == IP_STATE_OPENED))){ 
        	/* protocol family is not supported */
		DEBUG_TX("%s: Protocol family is not enabled (%d)!\n",
				card->devname, dst->sa_family);
	        return ENOPROTOOPT;
	}
#endif

	/* classify the packet before prepanding link-headers */
	IFQ_CLASSIFY(&ifp->if_snd, m0, dst->sa_family, &pktattr);

	if (card->wandev.state != WAN_CONNECTED){
       		IF_DROP(&ifp->if_snd);
		DEBUG_TX("%s: Drop the packet (Device is not connected)!\n", 
				card->devname);
		wan_skb_free(m0);	/*m_freem(m0);*/
	        error = ENETDOWN;
	}else{
		wan_spinlock_t	s;
		wan_spin_lock_irq(NULL, &s);

		IFQ_ENQUEUE(&ifp->if_snd, m0, &pktattr, error);
		if (error){
			DEBUG_TX("%s: Send queue is full!\n",
					card->devname);
			wan_spin_unlock_irq(NULL, &s);
			return error;	
		}

		if ((ifp->if_flags & IFF_OACTIVE) == 0){
			ppp_start(ifp);
		}
		wan_spin_unlock_irq(NULL, &s);
	}

	/* update statistics */
	/*ifp->if_obytes += m0->m_pkthdr.len;*/
	return error;
}

/*
 * ptpipe_start() - called by the system when a packet is added to the 
 *		    output queue for the interface and the device flags
 *		    indicate the device is not currently busy ( IFF_OACTIVE
 *                  is clear ).
 *                  We will try to transmit the packet. In case there's no
 *  		    available buffer - we will set IFF_OACTIVE, enable 
 *		    transmit interrupt and exit. And then the tx interrupt 
 *		    handler will call ptpipe_start() and once its done - it 
 *		    will reset the IFF_OACTIVE.
 */
static void
ppp_start(struct ifnet * ifp)
{
 	ppp_private_area_t*	ppp_priv_area = ifp->if_softc;
	sdla_t*	         	card = ppp_priv_area->common.card;
	ppp_buf_ctl_t       	txbuf;
	struct mbuf*		opkt = NULL;		/* the packet */
	register struct mbuf*	m = NULL;		/* current mbuf*/
	unsigned long 	 	data_addr;		/* where to copy data*/
	unsigned short	 	data_length;		/* data length */
	struct ip*			ip_pkt = NULL;
	wan_spinlock_t		s;		        /* saved int. state */

	wan_spin_lock_irq(NULL, &s);

	/* We're trying to drain the queue as much as possible here because */
	/* if we manage to get more than one packet into transmit queue     */
	/* ( line-down/line-up scenario, or heavy load ) and take only one  */
	/* at a time out we get an effective delay ...		      */
	/* 							      BF-0001 */ 

	while(!IFQ_IS_EMPTY(&ifp->if_snd)){

		IFQ_POLL(&ifp->if_snd, opkt);
		if (opkt == NULL) break;
#if 0
 /* !ALTQ */
    while (NULL != (opkt = ifp->if_snd.ifq_head)){	/* BF-0001 */ 
#endif

		/* calculate tx element address */
		card->hw_iface.peek(card->hw, card->u.p.txbuf_off, &txbuf, sizeof(ppp_buf_ctl_t));
		/* check if the buffer is available */
		if (txbuf.flag == 0x01) { 
			/* no buffers */
			DEBUG_TX("%s: TX buffers are full!!! Enabling TX INTR\n", 
					card->devname);
			/* mark an interface active */
			ifp->if_flags |= IFF_OACTIVE;

			ppp_priv_area->tick_counter = SYSTEM_TICKS;
			ppp_priv_area->if_send_stat.if_send_adptr_bfrs_full++;
			ppp_priv_area->if_send_stat.if_send_tx_int_enabled++;

			/* enable transmit interrupts */
			card->hw_iface.set_bit(card->hw, card->intr_perm_off, PPP_INTR_TXRDY);
			/* restore priority level and exit */
			wan_spin_unlock_irq(NULL, &s);
			return;
	        }
  
		/* 
		 * ok, now when we've got an available buffer we can transmit something
		 */
		/* initialize some of our stuff */
		data_addr 	= txbuf.buf.ptr;
		data_length	= 0;
	
		/* Calculate packet size */
		data_length = 0;
		for(m = opkt; m != NULL; m = m->m_next) 
			data_length += m->m_len;
		/* 
		 * Check that packet length is less than MAX_PACKET_SIZE.
		 * Copy tranmit packet to tx_data buffer.
		 */
		if (data_length > MAX_PACKET_SIZE){
			DEBUG_EVENT("%s: Tranmit packet is too long (%d)!\n",
					card->devname, data_length);
			/* dequeue packet from IP queue */
			IFQ_DEQUEUE(&ifp->if_snd, opkt);
			if (opkt){
				wan_skb_free(opkt);	/*m_freem(opkt);*/
			}
#if 0
 /* !ALTQ */

	    IF_DEQUEUE(&ifp->if_snd, opkt);
	    wan_skb_free(opkt);	/*m_freem(opkt);*/

#endif
			IFQ_INC_DROPS(&ifp->if_snd);
			continue;
		}
	
		data_length = 0;	
		for(m = opkt; m != NULL; m = m->m_next){
			bcopy(mtod(m, caddr_t), &card->tx_data[data_length], m->m_len);
			data_length += m->m_len;
		}

		ip_pkt = (struct ip*)&card->tx_data[0];
		card->hw_iface.poke(card->hw, data_addr, card->tx_data, data_length);

#ifdef ENABLE_IPV6
# ifdef __FreeBSD__
		txbuf->proto = (ip_pkt->ip_v == 0x4) ? 
					PROTOCOL_TYPE_IP : PROTOCOL_TYPE_IPv6;
# else
		txbuf.proto = (ip_pkt->ip_v == 0x4) ? 
					PROTOCOL_TYPE_IP : PROTOCOL_TYPE_IPv6;
		card->hw_iface.poke(card->hw, card->u.p.txbuf_off, 
					&txbuf, sizeof(txbuf));
# endif
#endif

		/* write data length to the transmit element - offset 1, length 2 */
		txbuf.length = data_length;
		card->hw_iface.poke(card->hw, card->u.p.txbuf_off, &txbuf, sizeof(txbuf)); 
		/* start transmission */
		txbuf.flag = 0x01;
		card->hw_iface.poke(card->hw, card->u.p.txbuf_off, &txbuf, sizeof(txbuf)); 

		/* report the packet to BPF if present and attached */
		wan_bpf_report(ifp, opkt, 1);

		/* dequeue packet from IP queue */
		IFQ_DEQUEUE(&ifp->if_snd, opkt);
		if (opkt){
			wan_skb_free(opkt);	/*m_freem(opkt);*/
		}
#if 0
 /* !ALTQ */
	IF_DEQUEUE(&ifp->if_snd, opkt);
	wan_skb_free(opkt);	/*m_freem(opkt);*/
#endif

		/* update statistics and free the packet */
		ifp->if_opackets++;
		ifp->if_obytes += data_length;

		/* Update transmit buffer control fields */
		card->u.p.txbuf_off += sizeof(ppp_buf_ctl_t);
		ppp_priv_area->if_send_stat.if_send_bfr_passed_to_adptr++;

		if (card->u.p.txbuf_off > card->u.p.txbuf_last_off){
			card->u.p.txbuf_off = card->u.p.txbuf_base_off;
		}

	} /* while there're packets in the queue */

	wan_spin_unlock_irq(NULL, &s);
	return;
}


/*=============================================================================
 * Store a UDP management packet for later processing.
 */
static int store_udp_mgmt_pkt(char udp_pkt_src, sdla_t* card,
                              struct mbuf *m, struct ifnet* ifp,
                              ppp_private_area_t* ppp_priv_area)
{
	int udp_pkt_stored = 0;
	
	if (!ppp_priv_area->udp_pkt_lgth && (m->m_len <= MAX_LGTH_UDP_MGNT_PKT)){
		ppp_priv_area->udp_pkt_lgth = m->m_len;
		ppp_priv_area->udp_pkt_src = udp_pkt_src;
       		bcopy(mtod(m, caddr_t), ppp_priv_area->udp_pkt_data, m->m_len);
		ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_UDP;
		ppp_priv_area->protocol = htons(ETHERTYPE_IP); 
		udp_pkt_stored = 1;
	}else{
		if (m->m_len > MAX_LGTH_UDP_MGNT_PKT){
			DEBUG_EVENT("%s: PIPEMON UDP request too long : %d\n",
					card->devname, m->m_len);
		}else{
			DEBUG_EVENT("%s: PIPEMON UDP request already pending\n",
					card->devname);
		}
		ppp_priv_area->udp_pkt_lgth = 0;
	}
	wan_skb_free(m);	/*m_freem(m);*/
	return(udp_pkt_stored);
}


#ifdef ENABLE_IPV6
static int reply_udp_ipv6( unsigned char *data, unsigned int mbox_len ) 
{
	unsigned short len, udp_length, temp, ip_length;
	int even_bound = 0;
	struct ip* 		ip4 = NULL;
	struct ip6_hdr* 	ip6 = NULL;
	struct udphdr* 	udp = NULL;
	wp_mgmt_t*		mgmt = NULL;

	if ((data[0] >> 4) == 6){
		len = sizeof(struct ip6_hdr);
	}else{
		len = sizeof(struct ip);
	}

	ip6 = (struct ip6_hdr*)data;
	ip4 = (struct ip*)data;
	udp = (struct udphdr*)&data[len];
	mgmt = (wan_mgmt_t*)&data[len + sizeof(struct udphdr)];

	/* Set length of packet */
	len +=  sizeof(struct udphdr)+ 
		sizeof(wan_mgmt_t)+
		sizeof(wan_cmd_t)+
		mbox_len;

	/* fill in UDP reply */
	mgmt->request_reply = UDPMGMT_REPLY; 
	
	/* fill in UDP length */
	udp_length = 	sizeof(struct udphdr)+ 
			sizeof(wan_mgmt_t)+
			sizeof(wan_cmd_t)+
			mbox_len; 
 
	/* put it on an even boundary */
	if (udp_length & 0x0001){ 
		udp_length += 1;
		len += 1;
		even_bound=1;
	} 

	/*temp = (udp_length<<8)|(udp_length>>8);
	  udp->uh_ulen = temp;*/
	udp->uh_ulen = htons(udp_length);

	/* swap UDP ports */
	temp = udp->uh_sport;
	udp->uh_sport = udp->uh_dport; 
	udp->uh_dport = temp;

	/* add UDP pseudo header */
	temp = 0x1100;
	*((unsigned short *)(data + len)) = temp;
	/*temp = (udp_length<<8)|(udp_length>>8);
	 *((unsigned short *)(data + 2)) = temp;*/
	*((unsigned short *)(data + len + 2)) = htons(udp_length);

	/* calculate UDP checksum */
	udp->uh_sum = 0;
	/*udp->uh_sum = wan_calc_checksum(&data[UDP_OFFSET],udp_length+UDP_OFFSET);*/
	/*udp->uh_sum = wan_calc_checksum(&data[UDP_OFFSET],udp_length+36);*/
	
	/* fill in IP length */
	if ((data[0] >> 4) == 6){
		struct in6_addr ip_temp;

	    	/*ip_length = udp_length + sizeof(struct ip6_hdr);*/
    		ip_length = udp_length;

	    	ip6->ip6_plen = htons(ip_length);

        	/* swap IP addresses */
	    	ip_temp = ip6->ip6_src;
    		ip6->ip6_src = ip6->ip6_dst;
	    	ip6->ip6_dst = ip_temp;
	}else{
		struct in_addr ip_temp;

		udp->uh_sum = wan_calc_checksum(&data[UDP_OFFSET],udp_length+UDP_OFFSET);
	
    		ip_length = udp_length + sizeof(struct ip);
	
    		ip4->ip_len = htons(ip_length);
	
    		/* swap IP addresses */
	    	ip_temp = ip4->ip_src;
    		ip4->ip_src = ip4->ip_dst;
	    	ip4->ip_dst = ip_temp;

    		/* fill in IP checksum */
	    	ip4->ip_sum = 0;
    		ip4->ip_sum = wan_calc_checksum(data,sizeof(struct ip));
	}

	return len;
} /* reply_udp */
#endif
/*
 ****************************************************************************
 *  If incoming is 0 (outgoing)- if the net numbers is ours make it 0
 *  if incoming is 1 - if the net number is 0 make it ours 
*/
#ifdef SUPPORT_IPX
static void 
switch_net_numbers(unsigned char *sendpacket, unsigned long network_number, unsigned char incoming)
{
    unsigned long pnetwork_number;

    pnetwork_number = (unsigned long)((sendpacket[6] << 24) + 
		      (sendpacket[7] << 16) + (sendpacket[8] << 8) + 
		      sendpacket[9]);

    if (!incoming){ 
	/* If the destination network number is ours, make it 0 */
	if (pnetwork_number == network_number){ 
	    sendpacket[6] = sendpacket[7] = sendpacket[8] = 
				 sendpacket[9] = 0x00;
	}
    }else{ 
	/* If the incoming network is 0, make it ours */
	if (pnetwork_number == 0){ 
	    sendpacket[6] = (unsigned char)(network_number >> 24);
  	    sendpacket[7] = (unsigned char)((network_number & 
						 0x00FF0000) >> 16);
	    sendpacket[8] = (unsigned char)((network_number & 
						 0x0000FF00) >> 8);
	    sendpacket[9] = (unsigned char)(network_number & 
						 0x000000FF);
	}
    }

    pnetwork_number = (unsigned long)((sendpacket[18] << 24) + 
		      (sendpacket[19] << 16) + (sendpacket[20] << 8) + 
		      sendpacket[21]);

    if (!incoming){ 
	/* If the source network is ours, make it 0 */
	if (pnetwork_number == network_number){
	    sendpacket[18] = sendpacket[19] = sendpacket[20] = 
					 sendpacket[21] = 0x00;
	}
    }else{ 
	/* If the source network is 0, make it ours */
	if (pnetwork_number == 0){ 
	    sendpacket[18] = (unsigned char)(network_number >> 24);
	    sendpacket[19] = (unsigned char)((network_number & 
						 0x00FF0000) >> 16);
	    sendpacket[20] = (unsigned char)((network_number & 
						 0x0000FF00) >> 8);
	    sendpacket[21] = (unsigned char)(network_number & 
						 0x000000FF);
	}
    }
} /* switch_net_numbers */
#endif


/****** Interrupt Handlers **************************************************/

/*
 * ============================================================================
 * PPP interrupt service routine.
 */
STATIC void wpp_isr(sdla_t* card)
{
	netdevice_t		*ifp = NULL;
	ppp_private_area_t 	*ppp_priv_area = NULL;
	unsigned char		intr_type = 0x00;
	char			ch;
	int 			i;

	card->in_isr = 1;
	
	++card->statistics.isr_entry;

	/* FIXME: Do we need this */
	card->force_enable_irq = 0;

	card->buff_int_mode_unbusy = 0;

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	if (ifp == NULL){
		goto wpp_isr_done;
	}
    	ppp_priv_area = wan_netif_priv(ifp);
	card->hw_iface.peek(card->hw, card->intr_type_off, &intr_type, sizeof(char));
	switch (intr_type){
	case PPP_INTR_RXRDY:	/* receive interrupt  0x01  (bit 0)*/
		++card->statistics.isr_rx;
		rx_intr(card);
		break;

	case PPP_INTR_TXRDY:	/* transmit interrupt  0x02 (bit 1)*/
        	/* 
		 * 1. Disable further tranmit interrupts.
		 * 2. Check the interface queue - if its not empty call ptpipe_start().
		 * 3. Reset the IFF_OACTIVE.
		 */
		++card->statistics.isr_tx;
		card->hw_iface.clear_bit(card->hw, 
					    card->intr_perm_off, 
					    PPP_INTR_TXRDY);
		/* interface is not active */
		if (ifp){
			ifp->if_flags &= ~IFF_OACTIVE;
		}
		/* Update tick_counter */
		if (ppp_priv_area){
			ppp_priv_area->tick_counter = SYSTEM_TICKS;
		}
		if (ifp && !IFQ_IS_EMPTY(&ifp->if_snd)){
			ppp_start(ifp);
		}
		card->buff_int_mode_unbusy = 1;
		break;

	case PPP_INTR_CMD:      /* interface command completed */
		++Intr_test_counter;
		++card->statistics.isr_intr_test;
		break;

	case PPP_INTR_MODEM:    /* modem status change (DCD, CTS) 0x04 (bit 2)*/
	case PPP_INTR_DISC:  	/* Data link disconnected 0x10  (bit 4)*/
	case PPP_INTR_OPEN:   	/* Data link open 0x20  (bit 5)*/
	case PPP_INTR_DROP_DTR:	/* DTR drop timeout expired  0x40 bit 6 */
		event_intr(card);
		break;

	case PPP_INTR_TIMER:
		timer_intr(card);
		break;	 

	default:	/* unexpected interrupt */
		++card->statistics.isr_spurious;
		DEBUG_EVENT("%s: ID Bytes = ",card->devname);
		for(i = 0; i < 8; i ++) {
			card->hw_iface.peek(card->hw, 
					    card->intr_type_off + 0x28 + i, 
					    &ch, 
					    sizeof(char)); 
			DEBUG_EVENT("0x%02x ", ch);
		}
		DEBUG_EVENT("\n");	
		break;
	}
wpp_isr_done:
	card->in_isr = 0;
	card->hw_iface.poke_byte(card->hw, card->intr_type_off, 0x00);
	return;
}

/*============================================================================
 * Receive interrupt handler.
 */
static void rx_intr(sdla_t *card)
{
	netdevice_t		*ifp = NULL;
	ppp_buf_ctl_t      	rxbuf;
	char			ch;
	ppp_private_area_t*	ppp_priv_area = NULL;
	unsigned long		addr = 0;
	struct mbuf*		ipkt = NULL; 
	int 			protocol = 0, 
				udp_type = 0;
	unsigned 		len;
	int 			i;
	int			isr = 0;
#ifdef ENABLE_IPV6
	sa_family_t		af = AF_INET;
	int 			err = 0;
	struct ifqueue*		ifq = NULL;
#endif

	card->hw_iface.peek(card->hw, card->rxmb_off, &rxbuf, sizeof(ppp_buf_ctl_t));
	if (rxbuf.flag != 0x01) { 
		DEBUG_EVENT("%s: corrupted Rx buffer @ 0x%lx, flag = 0x%02x!\n", 
				card->devname, card->rxmb_off, rxbuf.flag);
		DEBUG_EVENT("%s: ID Bytes = ",card->devname);
		for(i = 0; i < 8; i ++) {
			card->hw_iface.peek(card->hw, card->intr_type_off + 0x28 + i, &ch, 1);
			DEBUG_EVENT("0x%02x ", ch);
		}
		DEBUG_EVENT("\n");	
		
		++card->statistics.rx_intr_corrupt_rx_bfr;
		/*
 		 * 3.10.2000 RX_CORR
		 * This case not suppose to be, but if it is, disable interrupt
		 * for current router.
		 */
		DEBUG_EVENT("%s: Critical router failure...\n", card->devname);
		DEBUG_EVENT("Please contact yout Sangoma representation!!\n");
 		ppp_set_intr_mode(card, 0);	
		return;
	}

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	if (ifp == NULL){
		DEBUG_EVENT("%s: Internal Error (line no. %d)!\n", 
				card->devname, __LINE__);
		goto rx_intr_done;	
	} 
	ppp_priv_area	= wan_netif_priv(ifp);
	addr		= rxbuf.buf.ptr;
	len		= rxbuf.length;
    
	/*
	 * Check that packet len is less then MAX_PACKET_SIZE 
	 */
	if (len > MAX_PACKET_SIZE){
		DEBUG_EVENT("%s: Received packet length is too long!\n", 
				card->devname);
		goto rx_intr_done;	
	}
	/* 
	 * Now, we can start copying data into mbuf data area given 
	 * by mtod() macro. We copy m->m_len bytes. Note that since 
	 * the S508s buffers are cyclic we might have to copy the 
	 * packet in two pieces
	 */
	if (addr + len <= card->u.p.rx_top_off + 1){
        	/* copy iRn one piece */
	        card->hw_iface.peek(card->hw, addr, card->rx_data, len);
    		addr += len;
	}else{  
	    	/* copy in two pieces */
    		int piece1_len = card->u.p.rx_top_off + 1 - addr;
	    	int piece2_len = len - piece1_len;
  
    		card->hw_iface.peek(card->hw, addr, card->rx_data, piece1_len);
	    	card->hw_iface.peek(card->hw, card->u.p.rx_base_off, 
    				card->rx_data + piece1_len, piece2_len);
	    	addr = card->u.p.rx_base_off + piece2_len;
	}
   
	ipkt = sdla_get(ifp, card->rx_data, len);
   
	if (ipkt == NULL) { 
		DEBUG_EVENT("%s: no socket buffers available!\n", card->devname);
		++ppp_priv_area->rx_intr_stat.rx_intr_no_socket;
		ifp->if_iqdrops++;
		ifp->if_ierrors++;
		goto rx_intr_done;
	}

	/* Decapsulate packet */
	switch (rxbuf.proto){ 
	case PROTOCOL_TYPE_IP:
		protocol = htons(ETHERTYPE_IP);
#ifdef ENABLE_IPV6
		af = AF_INET;
#endif
		isr = NETISR_IP;
		break;

	case PROTOCOL_TYPE_IPX:
		protocol = htons(ETHERTYPE_IPX);
#ifdef ENABLE_IPV6
		af = AF_IPX;
#endif
		isr = NETISR_IPX;
		break;

#ifdef ENABLE_IPV6
	case PROTOCOL_TYPE_IPv6:
		protocol = htons(ETHERTYPE_IPV6);
		af = AF_INET6;
		isr = NETISR_IP6;
		break;
#endif
	}

	udp_type = udp_pkt_type(card, ipkt);
	if (udp_type == UDP_PTPIPE_TYPE){
		wan_bpf_report(ifp, ipkt, 1);
		/* Handle a UDP Request in Timer Interrupt */
		if (store_udp_mgmt_pkt(UDP_PKT_FRM_NETWORK, card, ipkt, ifp,
               	              			ppp_priv_area)) {
			card->hw_iface.set_bit(card->hw, card->intr_perm_off, PPP_INTR_TIMER);
		}
		++ppp_priv_area->rx_intr_stat.rx_intr_PIPE_request;
	}else{
#ifdef SUPPORT_IPX
		if (handle_IPXWAN(mtod(ipkt, caddr_t),card->devname, 
				 ppp_priv_area->enable_IPX, 
				 ppp_priv_area->network_number,
				 protocol)){ 
			/* Handle an IPXWAN packet */
			if (ppp_priv_area->enable_IPX){ 
				ppp_send(card, mtod(ipkt, caddr_t), ipkt->m_len, 
					htons(ETHERTYPE_IPX));
				wan_skb_free(m);	/*m_freem(m);*/
			}else{ 
				ifp->if_iqdrops++;
			}
		}else
#endif
        {
#if defined(__NetBSD__) || defined(__OpenBSD__) || (__FreeBSD_version < 501000)
		wan_spinlock_t	s;
#endif
		/* Ok, this is a regular packet - do what's said above */
		wan_bpf_report(ifp, ipkt, 1);

		/* 
		 * Schedule IP input routine. Note that it makes much more
		 * sence to check first if the protocol queue is full - and
		 * then to schedule the input routine, but i took a look at
		 * kernel sources for ether_input() 
		 * [ /usr/src/sys/net/if_ethersubr.c ] and they do it in 
		 * this order - so i figured they know better.
		 */
#ifdef ENABLE_IPV6
# if defined(__FreeBSD__) && (__FreeBSD_version >= 501000)
		netisr_queue(isr, ipkt);
# elif (defined __FreeBSD__) && (__FreeBSD_version > 400000)
		if ((err = family_enqueue(af, ipkt))){
			DEBUG_EVENT("%s: Failed to sned packet to IP stack (%d)!\n", 
					card->devname, err); 
			goto rx_intr_done;
		}
# else /* OpenBSD & FreeBSD-3.x */
		wan_spin_lock_irq(NULL, &s);
		schednetisr(isr);
		if (af == AF_INET){
			ifq = &ipintrq;
		}else{
			ifq = &ip6intrq;
		}
		if (IF_QFULL(ifq)){ 
			/* oh, no - IP queue is full - well - we'll try 
			 * again later 
			 */
			DEBUG_RX("%s: IP queue is full, drop packet!\n", 
						card->devname); 
			IF_DROP(&ifq);
			wan_skb_free(ipkt);	/*m_freem(ipkt);*/
			wan_spin_unlock_irq(NULL, &s);
			goto rx_intr_done;
		}
		/* ok enqueue the packet */
		IF_ENQUEUE(ifq, ipkt);
		wan_spin_unlock_irq(NULL, &s);
# endif	/* end of OpenBSD & FreeBSD-3.x */ 
#else /* IPv6 disabled */
# if (__FreeBSD_version >= 501000)
		netisr_queue(isr, ipkt);
# else
		wan_spin_lock_irq(NULL, &s);
		schednetisr(isr);
		if (IF_QFULL(&ipintrq)){
			/* oh, no - IP queue is full - well - we'll try 
			 * again later 
			 */
			DEBUG_RX("%s: IP queue is full, drop packet!\n", 
						card->devname); 
			IF_DROP(&ipintrq);
			wan_skb_free(ipkt);	/*m_freem(ipkt);*/
			wan_spin_unlock_irq(NULL, &s);
			goto rx_intr_done;
		}
		/* ok enqueue the packet */
		IF_ENQUEUE(&ipintrq, ipkt);
		wan_spin_unlock_irq(NULL, &s);
# endif
#endif
	 
		/* update interface statistics */
		ifp->if_ipackets++;
		ifp->if_ibytes += len;
		ppp_priv_area->rx_intr_stat.rx_intr_bfr_passed_to_stack++;
	}
   	}

rx_intr_done:

	/* Release buffer element and calculate a pointer to the next one */
	rxbuf.flag = 0x00;
	card->hw_iface.poke(card->hw, card->rxmb_off, &rxbuf, sizeof(rxbuf));
	card->rxmb_off += sizeof(ppp_buf_ctl_t);
	if (card->rxmb_off > card->u.p.rxbuf_last_off){
		card->rxmb_off = card->u.p.rxbuf_base_off;
	}
	return;
}


/*
 * ============================================================================
 * Receive event interrupt.
 */
void event_intr (sdla_t *card)
{
	netdevice_t 		*ifp = NULL;
	ppp_private_area_t	*ppp_priv_area = NULL;
	ppp_flags_t		flags;

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	WAN_ASSERT1(ifp == NULL);
	WAN_ASSERT1(wan_netif_priv(ifp) == NULL);
	ppp_priv_area = wan_netif_priv(ifp);

	card->hw_iface.peek(card->hw, card->flags_off, &flags, sizeof(ppp_flags_t));
	switch (flags.iflag){
	case PPP_INTR_MODEM:    /* modem status change (DCD, CTS) 0x04  (bit 2)*/
		if (IS_56K_CARD(card)) {
			FRONT_END_STATUS_STRUCT FE_status; 
			card->hw_iface.peek(card->hw, 0xF020, &FE_status, sizeof(FE_status));
			card->wandev.RR8_reg_56k = 
					FE_status.FE_U.stat_56k.RR8_56k;	
			card->wandev.RRA_reg_56k = 
					FE_status.FE_U.stat_56k.RRA_56k;	
			card->wandev.RRC_reg_56k = 
					FE_status.FE_U.stat_56k.RRC_56k;	
			sdla_56k_alarm(card, 0);
			FE_status.opp_flag = 0x01;
			card->hw_iface.poke(card->hw, 0xF020, &FE_status, sizeof(FRONT_END_STATUS_STRUCT));
			break;
		}

		if (IS_TE1_CARD(card)) {
			FRONT_END_STATUS_STRUCT FE_status; 
			card->hw_iface.peek(card->hw, 0xF020, &FE_status, sizeof(FRONT_END_STATUS_STRUCT));
			/* TE1 T1/E1 interrupt */
			sdla_te_intr(card);
			FE_status.opp_flag = 0x01;
			card->hw_iface.poke(card->hw, 0xF020, &FE_status, sizeof(FRONT_END_STATUS_STRUCT));
			break;
		}		
		DEBUG_EVENT("%s: Modem status: DCD=%s CTS=%s\n",
			     card->devname, DCD(flags.mstatus), CTS(flags.mstatus));
		break;

	case PPP_INTR_DISC:  	/* Data link disconnected 0x10  (bit 4) */
		DEBUG_EVENT("%s: Data link disconnected intr Cause %x\n",
					  card->devname, flags.disc_cause);
		if (flags.disc_cause &
		    (PPP_LOCAL_TERMINATION | PPP_DCD_CTS_DROP | PPP_REMOTE_TERMINATION))
		{
			if (card->u.p.ip_mode == WANOPT_PPP_PEER){ 
				Read_connection_info = 1;
				remove_route (card);
			}
			wanpipe_set_state(card, WAN_DISCONNECTED);
			ppp_priv_area->common.state = WAN_DISCONNECTED;
			WAN_DEBUG_START(card);
			show_disc_cause(card, flags.disc_cause);
			ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_PPP_EVENT;
			card->hw_iface.set_bit(card->hw, card->intr_perm_off, PPP_INTR_TIMER);
		}
		ifp->if_flags &= ~IFF_UP;
		break;

	case PPP_INTR_OPEN:   	/* Data link open 0x20  (bit 5)*/
#ifdef ENABLE_IPV6
		DEBUG_EVENT("%s: PPP Link Open, LCP=%s IP=%s IPv6=%s\n",
			   card->devname, LCP(flags.lcp_state), IP(flags.ip_state), 
			   IP(flags.ipv6_state));
#else
		DEBUG_EVENT("%s: PPP Link Open, LCP=%s IP=%s\n",
			   card->devname, LCP(flags.lcp_state), IP(flags.ip_state));
#endif
		/* Protocol is opened - mark interface*/
		ifp->if_flags |= IFF_RUNNING;
		/* reset disconnect cause */
        	/*
	         * ok, this is a hack - what happens is that sometimes when we bring 
		 * the interface up a disconnect interrupt occures which resets 
		 * the IFF_UP flag. The protocol-open interrupt always occures later on
		 * so we can fix it here. On second thought - the only way we can bring 
		 * up the interface which was down is through here - so i'll just be 
		 * always setting IFF_UPO flag here ;-)
		 */
		ifp->if_flags |= IFF_UP;

		/* update the up time ... */
		wan_getcurrenttime(&ppp_priv_area->router_up_time, NULL);

	        /* 
        	 * also from this call the interface begins "intelligent" existance so
		 * we would like to activate the watchdog routine that will monitor the
		 * protocol state and update the interface state accordingly.
		 */
		ifp->if_timer = 1; 	/* call it in a second */

		if (flags.lcp_state == LCP_STATE_OPENED && 
        	    (flags.ip_state == IP_STATE_OPENED ||
#ifdef ENABLE_IPV6
	             flags.ipv6_state == IP_STATE_OPENED || 
#endif
        	     flags.ipx_state == IP_STATE_OPENED))
		{
			/* Initialize the polling timer and set the state
			 * to WAN_CONNNECTED 
			 */
			card->state_tick = SYSTEM_TICKS;
			wanpipe_set_state(card, WAN_CONNECTED);
			ppp_priv_area->common.state = WAN_CONNECTED;
			ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_PPP_EVENT;
			card->hw_iface.set_bit(card->hw, card->intr_perm_off, PPP_INTR_TIMER);
		}
		break;

	case PPP_INTR_DROP_DTR:	/* DTR drop timeout expired  0x40 bit 6 */
		DEBUG_EVENT("%s: DTR Drop Timeout Interrrupt \n", card->devname); 
		if (card->u.p.ip_mode == WANOPT_PPP_PEER){ 
			Read_connection_info = 1;
			remove_route (card);
		}
		wanpipe_set_state(card, WAN_DISCONNECTED);
	    	ppp_priv_area->common.state = WAN_DISCONNECTED;
		WAN_DEBUG_START(card);
		show_disc_cause(card, flags.disc_cause);
		ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_PPP_EVENT;
		card->hw_iface.set_bit(card->hw, card->intr_perm_off, PPP_INTR_TIMER);
		break;

	default:
		DEBUG_EVENT("%s: Error, Invalid PPP Event\n",card->devname);
	}
}


/*
 * ============================================================================
 * Timer interrupt.
 */
void timer_intr (sdla_t *card)
{
	netdevice_t		*ifp = NULL;
	ppp_private_area_t* ppp_priv_area = NULL;

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	WAN_ASSERT1(ifp == NULL);
	WAN_ASSERT1(wan_netif_priv(ifp) == NULL);
	ppp_priv_area = wan_netif_priv(ifp);

	/* Interrupt test */
#ifndef PPP_OLD_INTR_TEST
	if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_TEST){
		Intr_test_counter++;
		if (Intr_test_counter >= MAX_NEW_INTR_TEST_COUNTER){
			DEBUG_EVENT("%s: Interrupt Test Passed, Counter: %d\n",
					card->devname, Intr_test_counter);
			ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_TEST;
			ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_CONFIG;
		}
	}
#endif 

	if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_CONFIG){
		ppp_config(ifp);
		ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_CONFIG;
	}

	/* Update statistics */
	if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_UPDATE){
		ppp_err_stats_t err_stats;
		ppp508_stats_t	stats;
		if (!ppp_get_err_stats(card, &err_stats)){
			ifp->if_ierrors	= err_stats.rx_overrun +
					  err_stats.rx_bad_crc +
					  err_stats.rx_abort +
					  err_stats.rx_lost;
			ifp->if_oerrors	= err_stats.tx_abort;
		}	
		if (!ppp_get_stats(card, &stats)){
			ifp->if_ierrors += stats.rx_bad_len;
		}
		if (!(--ppp_priv_area->update_comms_stats)){
			ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE;
		}
	}

	/* TE1 T1/E1 timer interrupt */
	if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_TE){
		sdla_te_polling(card);
		ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_TE;
	}

	/* PPIPEMON UDP request */
	if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_UDP){
		process_udp_mgmt_pkt(card,ifp, ppp_priv_area);
		ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UDP;
	}

	/* PPP Event */
	if (ppp_priv_area->timer_int_enabled & TMR_INT_ENABLED_PPP_EVENT){
		switch(card->wandev.state){
		case WAN_DISCONNECTED:
			poll_disconnected(card);
			break;
	
		case WAN_CONNECTING:	
			/* If the state is CONNECTING, it means that communicatins were
		 	 * enabled. When the remote side enables its comminication we
	 		 * should get an interrupt PPP_INTR_OPEN, thus turn off polling 
			 */
			/* Turn off the timer interrupt */
			ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_PPP_EVENT;
			break;
	
		case WAN_CONNECTED:
			/* If state is connected and we are in PEER mode 
			 * poll for an IP address which will be provided by remote end.
			 */
			if (card->u.p.ip_mode == WANOPT_PPP_PEER && Read_connection_info){
		    		card->state_tick = SYSTEM_TICKS;
				if (!read_connection_info (card)){
					ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_PROCESS_ROUTE;
					ifp->if_timer = 1; 
				}
			}else{
				ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_PPP_EVENT;
			}
			break;

		default:
			/* If we are using Static IP,no need to poll for
			 * an IP address. 
			 */
			ppp_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_PPP_EVENT;
			break;
		}	

	}/* End of PPP_EVENT */

	/* Only disable the timer interrupt if there are no udp, statistic */
	/* updates or events pending */
	if (!ppp_priv_area->timer_int_enabled) { 
		card->hw_iface.clear_bit(card->hw, card->intr_perm_off, PPP_INTR_TIMER);
	}
}


#ifdef SUPPPORT_IPX
static int 
handle_IPXWAN(unsigned char *sendpacket, char *devname, unsigned char enable_IPX, unsigned long network_number, unsigned short proto)
{
    int i;

    if (proto == htons(ETHERTYPE_IPX)){ 
	/* It's an IPX packet */
	if (!enable_IPX){
	    /* Return 1 so we don't pass it up the stack. */
	    return 1;
	}
    }else{ 
	/* It's not IPX so pass it up the stack. */
	return 0;
    }

    if (sendpacket[16] == 0x90 && sendpacket[17] == 0x04){
	/* It's IPXWAN */
	if (sendpacket[2] == 0x02 && sendpacket[34] == 0x00){
	    /* It's a timer request packet */ 
	    DEBUG_EVENT("%s: Received IPXWAN Timer Request packet\n", devname);

	    /* Go through the routing options and answer no to every */
	    /* option except Unnumbered RIP/SAP */
	    for(i = 41; sendpacket[i] == 0x00; i += 5){
		/* 0x02 is the option for Unnumbered RIP/SAP */
		if (sendpacket[i + 4] != 0x02){
		    sendpacket[i + 1] = 0;
		}
	    }

	    /* Skip over the extended Node ID option */
	    if (sendpacket[i] == 0x04){
		i += 8;
	    }

	    /* We also want to turn off all header compression opt. */
	    for(; sendpacket[i] == 0x80 ;){
		sendpacket[i + 1] = 0;
		i += (sendpacket[i + 2] << 8) + (sendpacket[i + 3]) + 4;
	    }

	    /* Set the packet type to timer response */
	    sendpacket[34] = 0x01;

	    DEBUG_EVENT("%s: Sending IPXWAN Timer Response\n", devname);
	}else
	    if (sendpacket[34] == 0x02){
		/* This is an information request packet */
		DEBUG_EVENT("%s: Received IPXWAN Information Request packet\n",
								devname);

		/* Set the packet type to information response */
		sendpacket[34] = 0x03;

		/* Set the router name */
		sendpacket[51] = 'P';
		sendpacket[52] = 'T';
		sendpacket[53] = 'P';
		sendpacket[54] = 'I';
		sendpacket[55] = 'P';
		sendpacket[56] = 'E';
		sendpacket[57] = '-';
		sendpacket[58] = CVHexToAscii(network_number >> 28);
		sendpacket[59] = 
			CVHexToAscii((network_number & 0x0F000000)>> 24);
		sendpacket[60] = 
			CVHexToAscii((network_number & 0x00F00000)>> 20);
		sendpacket[61] = 
			CVHexToAscii((network_number & 0x000F0000)>> 16);
		sendpacket[62] = 
			CVHexToAscii((network_number & 0x0000F000)>> 12);
		sendpacket[63] = 
			CVHexToAscii((network_number & 0x00000F00)>> 8);
		sendpacket[64] = 
			CVHexToAscii((network_number & 0x000000F0)>> 4);
		sendpacket[65] = CVHexToAscii(network_number & 0x0000000F);
		for(i = 66; i < 99; i+= 1)
		    sendpacket[i] = 0;

		DEBUG_EVENT("%s: Sending IPXWAN Information Response packet\n",
								devname);
	    }else{
		DEBUG_EVENT("%s: Unknown IPXWAN packet!\n",devname);
		return 0;
	    }

        /* Set the WNodeID to our network address */
        sendpacket[35] = (unsigned char)(network_number >> 24);
	sendpacket[36] = (unsigned char)((network_number & 0x00FF0000) >> 16);
    	sendpacket[38] = (unsigned char)((network_number & 0x0000FF00) >> 8);
        sendpacket[38] = (unsigned char)(network_number & 0x000000FF);

    	return 1;
    }else{ 
        /* If we get here's its an IPX-data packet, so it'll get passed 
         * up the stack. 
         */

	/* switch the network numbers */
	switch_net_numbers(sendpacket, network_number, 1);	
	return 0;
    }
}
#endif

/****** Background Polling Routines  ****************************************/

/* All polling functions are invoked by the TIMER interrupt in the wpp_isr 
 * routine.  
 */

/*============================================================================
 * Monitor active link phase.
 */
static void process_route (netdevice_t* ifp)
{
	ppp_private_area_t*	ppp_priv_area = wan_netif_priv(ifp);
	sdla_t*			card = ppp_priv_area->common.card;
	ppp_flags_t		flags;

	card->hw_iface.peek(card->hw, card->flags_off + offsetof(ppp_flags_t, ip_state), 
			 &flags.ip_state, 
			 sizeof(flags.ip_state));
	if ((card->u.p.ip_mode == WANOPT_PPP_PEER) &&
        	(Read_connection_info && flags.ip_state == 0x09))
	{ 
	        DEBUG_EVENT("%s: IPCP State Opened.\n", card->devname);
        	read_info(card);
	        Read_connection_info = 0;
	}
	return;
}


/*============================================================================
 * Monitor physical link disconnected phase.
 *  o if interface is up and the hold-down timeout has expired, then retry
 *    connection.
 */
static void poll_disconnected(sdla_t *card)
{
	netdevice_t		*ifp;
	ppp_private_area_t	*ppp_priv_area = NULL;

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	WAN_ASSERT1(ifp == NULL);
	WAN_ASSERT1(wan_netif_priv(ifp) == NULL);
	ppp_priv_area = wan_netif_priv(ifp);

	if ((SYSTEM_TICKS - card->state_tick) > HOLD_DOWN_TIME){ 
		wanpipe_set_state(card, WAN_CONNECTING);
		ppp_priv_area->common.state = WAN_CONNECTING;
		if (ppp_comm_enable(card) == WAN_CMD_OK)
			init_ppp_tx_rx_buff(card);
	}
}

/****** Miscellaneous Functions *********************************************/

/*============================================================================
 * Configure S508 adapter.
 */
static int 
config508(ppp_private_area_t *ppp_priv_area, sdla_t *card)
{
	ppp508_conf_t	cfg;

	/* Prepare PPP configuration structure */
	bzero(&cfg, sizeof(ppp508_conf_t));

	if (card->wandev.clocking){
		cfg.line_speed = card->wandev.bps;
	}

	if (card->wandev.interface == WANOPT_RS232){
		cfg.conf_flags |= INTERFACE_LEVEL_RS232;
	}

	/*send Configure-Request packets forever*/
	cfg.conf_flags 	|= DONT_TERMINATE_LNK_MAX_CONFIG; 
	cfg.txbuf_percent	= PERCENT_TX_BUFF;	/* % of Tx bufs */
	cfg.mtu_local	= card->wandev.mtu;
	cfg.mtu_remote	= card->wandev.mtu;                  /*    Default   */
	cfg.restart_tmr	= TIME_BETWEEN_CONF_REQ;  	     /*    30 = 3sec */
	cfg.auth_rsrt_tmr	= TIME_BETWEEN_PAP_CHAP_REQ;         /*    30 = 3sec */
	cfg.auth_wait_tmr	= WAIT_PAP_CHAP_WITHOUT_REPLY;       /*   300 = 30s  */
	cfg.mdm_fail_tmr	= WAIT_AFTER_DCD_CTS_LOW;            /*     5 = 0.5s */
	cfg.dtr_drop_tmr	= TIME_DCD_CTS_LOW_AFTER_LNK_DOWN;   /*    10 = 1s   */
	cfg.connect_tmout	= WAIT_DCD_HIGH_AFTER_ENABLE_COMM;   /*   900 = 90s  */
	cfg.conf_retry	= MAX_CONF_REQ_WITHOUT_REPLY;        /*    10 = 1s   */
	cfg.term_retry	= MAX_TERM_REQ_WITHOUT_REPLY;	     /*     2 times  */
	cfg.fail_retry	= NUM_CONF_NAK_WITHOUT_REPLY;        /*     5 times  */
	cfg.auth_retry	= NUM_AUTH_REQ_WITHOUT_REPLY;        /*     10 times */  
	if (!card->u.p.authenticator){ 
		DEBUG_EVENT("%s: Device is not configured as an authenticator\n",
						card->devname);
		cfg.auth_options = NO_AUTHENTICATION;
	}else{
		DEBUG_EVENT("%s: Device is configured as an authenticator\n", 
						card->devname);
		cfg.auth_options = INBOUND_AUTH;
	}	
	if (ppp_priv_area->pap == WANOPT_YES){
		cfg.auth_options |=PAP_AUTH;
		DEBUG_EVENT("%s: Pap enabled\n", card->devname);
	}
	if (ppp_priv_area->chap == WANOPT_YES){
		cfg.auth_options |= CHAP_AUTH;
		DEBUG_EVENT("%s: Chap enabled\n", card->devname);
	}

	if (ppp_priv_area->enable_IPX == WANOPT_YES){
		cfg.ipx_options		= ENABLE_IPX | ROUTING_PROT_DEFAULT;
	}else{
		cfg.ipx_options 	= DISABLE_IPX;
	}

	/* Local IP address */
	cfg.ip_local = ppp_priv_area->ip_local;
	/* Destination IP address */
	cfg.ip_remote = ppp_priv_area->ip_remote;
	switch (card->u.p.ip_mode){ 
	case WANOPT_PPP_STATIC:
		DEBUG_EVENT("%s: PPP IP mode: STATIC.\n",
				card->devname);
		cfg.ip_options		= L_AND_R_IP_NO_ASSIG | ENABLE_IP;

		DEBUG_EVENT("%s: Local  %s\n", 
					card->devname, in_ntoa(cfg.ip_local));
		DEBUG_EVENT("%s: Remote %s\n", 
					card->devname, in_ntoa(cfg.ip_remote));
		break;

	case WANOPT_PPP_HOST:
		DEBUG_EVENT("%s: PPP IP mode: HOST.\n",
				card->devname);
		cfg.ip_options		= L_IP_LOCAL_ASSIG | 
					  R_IP_LOCAL_ASSIG | ENABLE_IP;
		break;

	case WANOPT_PPP_PEER:
		DEBUG_EVENT("%s: PPP IP mode: PEER.\n",
					card->devname);
		cfg.ip_options		= L_IP_REMOTE_ASSIG | 
					  R_IP_REMOTE_ASSIG | 
					  ENABLE_IP;
		break;
	}
	return ppp_configure(card, &cfg);
}

#ifdef ENABLE_IPV6
static int ppp_get_ip_cfg(ppp_private_area_t *ppp_priv_area, ppp508_conf_t* cfg, struct ifaddr* ifa)
{
    sdla_t* 			card = (sdla_t*)ppp_priv_area->card;
    struct ifnet* 		ifp = NULL;
    struct sockaddr_in*		saddr = NULL;
    sa_family_t 		af = AF_INET;
    int 			err = 0;

    ifp = card->wandev.dev;
    af = (ifa->ifa_addr->sa_family == AF_INET) ? AF_INET : AF_INET6;

    if (af == AF_INET && cfg == NULL){
	DEBUG_EVENT("%s: Invalid arguments\n", card->devname);
	return -EINVAL;
    }
    switch (card->u.p.ip_mode){ 
    case WANOPT_PPP_STATIC:
	DEBUG_EVENT("%s: PPP %s mode: STATIC.\n",
				card->devname, 
				(af == AF_INET) ? "IP" : "IPv6");
	if (af == AF_INET){
	    cfg->ip_options		= L_AND_R_IP_NO_ASSIG | ENABLE_IP;
	    saddr = (struct sockaddr_in*)ifa->ifa_addr;
	    if (saddr)
	    	cfg->ip_local = saddr->sin_addr.s_addr;
	    /* Destination IP address */
	    saddr = (struct sockaddr_in*)ifa->ifa_dstaddr;
	    if (saddr)
	    	cfg->ip_remote = saddr->sin_addr.s_addr;
	    DEBUG_EVENT("%s: Local  %s\n", 
				card->devname, in_ntoa(cfg->ip_local));
	    DEBUG_EVENT("%s: Remote %s\n", 
				card->devname, in_ntoa(cfg->ip_remote));
	}else{
	    struct in6_ifaddr* ifa6 = (struct in6_ifaddr*)ifa;
	
    	    bcopy(&ifa6->ia_addr.sin6_addr, &ppp_priv_area->ip6_local, sizeof(struct in6_addr));
    	    bcopy(&ifa6->ia_dstaddr.sin6_addr, &ppp_priv_area->ip6_remote, sizeof(struct in6_addr));

	    DEBUG_EVENT("%s: Local  %s\n", 
			card->devname, ip6_sprintf(&ppp_priv_area->ip6_local));
	    DEBUG_EVENT("%s: Remote %s\n", 
			card->devname, ip6_sprintf(&ppp_priv_area->ip6_remote));
	}
	break;

    case WANOPT_PPP_HOST:
	DEBUG_EVENT("%s: PPP %s mode: HOST.\n",
			card->devname, (af == AF_INET) ? "IP" : "IPv6");
	if (af == AF_INET){
	    cfg->ip_options	= L_IP_LOCAL_ASSIG | 
	     			  R_IP_LOCAL_ASSIG | ENABLE_IP;
	    /* Local IP address */
	    saddr = (struct sockaddr_in*)ifa->ifa_addr;
	    if (saddr)
	        cfg->ip_local = saddr->sin_addr.s_addr;
	    /* Destination IP address */
	    saddr = (struct sockaddr_in*)ifa->ifa_dstaddr;
	    if (saddr)
	        cfg->ip_local = saddr->sin_addr.s_addr;
	}else{
	    err = -EINVAL;
	}
	break;

    case WANOPT_PPP_PEER:
	DEBUG_EVENT("%s: PPP %s mode: PEER.\n",
			card->devname, (af == AF_INET) ? "IP" : "IPv6");
	if (af == AF_INET){
	    cfg->ip_options	= L_IP_REMOTE_ASSIG | 
				  R_IP_REMOTE_ASSIG | 
				  ENABLE_IP;
	    cfg->ip_local	= 0x00;
	    cfg->ip_remote	= 0x00;
	}else{
	    err = -EINVAL;
	}
	break;
    }


    return err;
}
#endif

/*============================================================================
 * Show disconnection cause.
 */
static void show_disc_cause(sdla_t *card, unsigned cause)
{
	if (cause & 0x0802)
		DEBUG_EVENT("%s: link terminated by peer\n", card->devname);
	else if (cause & 0x0004)
        	DEBUG_EVENT("%s: link terminated by user\n", card->devname);
	else if (cause & 0x0008) 
		DEBUG_EVENT("%s: authentication failed\n", card->devname);
	else if (cause & 0x0010) 
		DEBUG_EVENT("%s: authentication protocol negotiation failed\n", 
							card->devname);
	else if (cause & 0x0020) 
		DEBUG_EVENT("%s: peer's request for authentication rejected\n",
							card->devname);
	else if (cause & 0x0040) 
		DEBUG_EVENT("%s: MRU option rejected by peer\n", 
							card->devname);
	else if (cause & 0x0080) 
		DEBUG_EVENT("%s: peer's MRU was too small\n", 
							card->devname);
	else if (cause & 0x0100) 
		DEBUG_EVENT("%s: failed to negotiate peer's LCP options\n",
							card->devname);
	else if (cause & 0x0200) 
		DEBUG_EVENT("%s: failed to negotiate peer's IPCP options\n",
							card->devname);
	else if (cause & 0x0400) 
		DEBUG_EVENT("%s: failed to negotiate peer's IPXCP options\n",
							card->devname);
}

/*
 * Execute PPP udp commands
 */
static int ppp_udp_ioctl(struct ifnet* ifp, caddr_t data)
{
    	ppp_private_area_t*	ppp_priv_area = ifp->if_softc; 
    	sdla_t*			card = ppp_priv_area->common.card;
	struct ifreq*		ifr = (struct ifreq*)data;
	wan_udp_hdr_t		udp_hdr;
	int 			error = 0;

	if (copyin((void*)ifr->ifr_data, (void*)&udp_hdr, sizeof(wan_udp_hdr_t)) < 0){
		DEBUG_EVENT("%s: Failed to copy data from user to kernel space\n",
				card->devname);
		return -EFAULT;
	}

	error = process_udp_cmd(ifp, &udp_hdr);	

	if (copyout((void*)&udp_hdr, (void*)ifr->ifr_data, sizeof(wan_udp_hdr_t)) < 0){
		DEBUG_EVENT("%s: Failed to copy data from kernel to user space\n",
				card->devname);
		return -EFAULT;
	}
	return 0;
}


/*
 ***************************************************************************
 *
 *
 */

static int process_udp_cmd(struct ifnet* ifp, wan_udp_hdr_t* udp_hdr)
{
    	ppp_private_area_t*	ppp_priv_area = ifp->if_softc; 
    	sdla_t*			card = ppp_priv_area->common.card;
	trace_element_t 	trace_element;
    	unsigned char 		buf2[5];
    	unsigned int 		frames;
    	unsigned short 		data_length = 0, 
				buffer_length = 0, 
				real_len = 0;
    	unsigned long 		data_ptr;
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 			err = 0;

	/* Initialize the trace element */

	switch (udp_hdr->wan_udphdr_command){
   	case PPIPE_ENABLE_TRACING:
	    if (!card->TracingEnabled){ 
    		/* OPERATE_DATALINE_MONITOR */
		WAN_MBOX_INIT(mbox);
    		mbox->wan_command = PPP_DATALINE_MONITOR;
    		mbox->wan_data_len = 0x01;
   		mbox->wan_data[0] = udp_hdr->wan_udphdr_data[0];
    		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
		if (err != WAN_CMD_OK) { 
		    ppp_error(card, err, mbox);
        	    card->TracingEnabled = 0;
		    /* set the return code */
	       	    udp_hdr->wan_udphdr_return_code = mbox->wan_return_code;
	    	    udp_hdr->wan_udphdr_data_len = 0;
        	    break;
		}
		card->hw_iface.peek(card->hw, 0xC000, &buf2, 2);

		ppp_priv_area->curr_trace_addr = 0;
   		bcopy(&buf2, &ppp_priv_area->curr_trace_addr, 2);
    		ppp_priv_area->start_trace_addr = 
					ppp_priv_area->curr_trace_addr;
		ppp_priv_area->end_trace_addr = 
				ppp_priv_area->start_trace_addr + END_OFFSET;
		    	
		/* MAX_SEND_BUFFER_SIZE - 28 (IP header) 
		 *  - 32 (ppipemon CBLOCK) 
		 */
	    	available_buffer_space = MAX_LGTH_UDP_MGNT_PKT - 
					 sizeof(struct ip)-
					 sizeof(struct udphdr)-
					 sizeof(wan_mgmt_t)-
					 sizeof(wan_cmd_t); 
       	    }
            udp_hdr->wan_udphdr_return_code = 0;
	    udp_hdr->wan_udphdr_data_len = 0;
	    card->TracingEnabled = 1;
	    break;
	  
	case PPIPE_DISABLE_TRACING:
	    if (card->TracingEnabled){ 

		/* OPERATE_DATALINE_MONITOR */
		WAN_MBOX_INIT(mbox);
  		mbox->wan_command = 0x33;
 		mbox->wan_data_len = 1;
		mbox->wan_data[0] = 0x00;
    		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	    } 

	    /*set return code*/
	    udp_hdr->wan_udphdr_return_code = 0;
	    udp_hdr->wan_udphdr_data_len = 0;
	    card->TracingEnabled = 0;
	    break;

	case PPIPE_GET_TRACE_INFO:
	    if (!card->TracingEnabled){ 
	  	/* set return code */
		udp_hdr->wan_udphdr_return_code = 1;
		udp_hdr->wan_udphdr_data_len = 0;
	    }		    

	    buffer_length = 0;

	    /* frames < 62, where 62 is the number of trace
	     * information elements.  There is in total 496
	     * bytes of space and each trace information
	     * element is 8 bytes. 
	     */
	    for (frames=0; frames<62; frames++){ 
		wan_trace_pkt_t *trace_pkt = 
			(wan_trace_pkt_t*)&udp_hdr->wan_udphdr_data[buffer_length];

		/* Read the whole trace packet */
		card->hw_iface.peek(card->hw, ppp_priv_area->curr_trace_addr, 
				  &trace_element, sizeof(trace_element_t));
	
		/* no data on board so exit */
		if (trace_element.opp_flag == 0x00) 
		    break;
	      
		data_ptr = trace_element.trace_data_ptr;

		/* See if there is actual data on the trace buffer */
		if (data_ptr){
		    data_length = trace_element.trace_length;
		}else{
		    data_length = 0;
		    udp_hdr->wan_udphdr_data[0] |= 0x02;
		}

		/* FIXME: Do we need this check */
		if ((available_buffer_space - buffer_length) < 
		    (sizeof(trace_element_t)+1))
		{
		    /* indicate we have more frames 
		     * on board and exit 
		     */
		    udp_hdr->wan_udphdr_data[0] |= 0x02;
		    break;
		}
	
		trace_pkt->status = trace_element.trace_type;
		trace_pkt->time_stamp = trace_element.trace_time_stamp;
		trace_pkt->real_length = trace_element.trace_length;

		real_len = trace_element.trace_length;	
	
		if (data_ptr == 0){
		    trace_pkt->data_avail = 0x00;
		}else{
		    /* we can take it next time */
		    if ((available_buffer_space - buffer_length) <
				(real_len + sizeof(wan_trace_pkt_t)))
		    {
			udp_hdr->wan_udphdr_data[0] |= 0x02;
			break;
		    } 
		    trace_pkt->data_avail = 0x01;
	
		    /* get the data */
		    card->hw_iface.peek(card->hw, data_ptr, &trace_pkt->data, real_len);
		}	
		/* zero the opp flag to 
		 * show we got the frame 
		 */
		card->hw_iface.poke_byte(card->hw, ppp_priv_area->curr_trace_addr, 0x00);

		/* now move onto the next 
		 * frame 
		 */
		ppp_priv_area->curr_trace_addr += sizeof(trace_element_t);

		/* check if we passed the last address */
		if (ppp_priv_area->curr_trace_addr >= 
		  				  ppp_priv_area->end_trace_addr){
		    ppp_priv_area->curr_trace_addr = 
				ppp_priv_area->start_trace_addr;
		}
 
		/* update buffer length and make sure its even */ 
		if (trace_pkt->data_avail == 0x01){ 
		    buffer_length += real_len - 1;
		}
 
		/* for the header */
		buffer_length += sizeof(wan_trace_pkt_t);
	    }

	    /* ok now set the total number of frames passed
	     * in the high 5 bits 
	     */
	    udp_hdr->wan_udphdr_data[0] |= (frames << 2);

	    /* set the data length */
	    udp_hdr->wan_udphdr_data_len = buffer_length;

	    /* set return code */
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

	case PPIPE_GET_IBA_DATA:
	    card->hw_iface.peek(card->hw, 0xF003, &udp_hdr->wan_udphdr_data[0], 0x09);
	        
	    /* set the length of the data */
	    udp_hdr->wan_udphdr_data_len = 0x09;

	    /* set return code */
	    udp_hdr->wan_udphdr_return_code = 0x00;
	    break;

	case PPIPE_KILL_BOARD:
	    break;

	case PPIPE_FT1_READ_STATUS:
	    card->hw_iface.peek(card->hw, 0xF020, &udp_hdr->wan_udphdr_data[0], 2);
	    udp_hdr->wan_udphdr_data_len = 2;
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

	case PPIPE_FLUSH_DRIVER_STATS:   
	    init_ppp_priv_struct(ppp_priv_area);
	    init_global_statistics(card);
	    udp_hdr->wan_udphdr_data_len = 0;
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

	case PPIPE_ROUTER_UP_TIME:
	    wan_getcurrenttime(&ppp_priv_area->router_up_time, NULL);
	    ppp_priv_area->router_up_time -= ppp_priv_area->router_start_time; 
	    *(unsigned long *)&udp_hdr->wan_udphdr_data[0] =
					ppp_priv_area->router_up_time;
	    udp_hdr->wan_udphdr_data_len = sizeof(unsigned long);
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

   	case FT1_MONITOR_STATUS_CTRL:
	    /* Enable FT1 MONITOR STATUS */
            if (udp_hdr->wan_udphdr_data[0] == 1){ 
		if (rCount++ != 0){ 
		    udp_hdr->wan_udphdr_return_code = 0;
	    	    udp_hdr->wan_udphdr_data_len = 1;
		    break;
    		}	
	    }

      	    /* Disable FT1 MONITOR STATUS */
      	    if (udp_hdr->wan_udphdr_data[0] == 0){ 
		if (--rCount != 0){ 
		    udp_hdr->wan_udphdr_return_code = 0;
	    	    udp_hdr->wan_udphdr_data_len = 1;
		    break;
   	    	} 
      	    } 
	    /* We should execute FT1 command now */
	    goto udp_dflt_cmd;
	    break;	

	case PPIPE_DRIVER_STAT_IFSEND:
	    bcopy(&ppp_priv_area->if_send_stat, &udp_hdr->wan_udphdr_data[0],   
						sizeof(if_send_stat_t));

	    udp_hdr->wan_udphdr_return_code = 0;
	    udp_hdr->wan_udphdr_data_len = sizeof(if_send_stat_t);
	    break;

	case PPIPE_DRIVER_STAT_INTR:
	    bcopy(&card->statistics, &udp_hdr->wan_udphdr_data[0],  
				sizeof(global_stats_t));
            bcopy(&ppp_priv_area->rx_intr_stat,
	    	  		&udp_hdr->wan_udphdr_data[sizeof(global_stats_t)],
				sizeof(rx_intr_stat_t));
	    udp_hdr->wan_udphdr_return_code = 0;
	    udp_hdr->wan_udphdr_data_len = sizeof(global_stats_t)+
			             sizeof(rx_intr_stat_t);
	    break;

	case PPIPE_DRIVER_STAT_GEN:
	    bcopy(&ppp_priv_area->pipe_mgmt_stat, &udp_hdr->wan_udphdr_data[0], 
					sizeof(pipe_mgmt_stat_t));
	    bcopy(&card->statistics,
	          &udp_hdr->wan_udphdr_data[sizeof(pipe_mgmt_stat_t)],
		  sizeof(global_stats_t)); 
	    udp_hdr->wan_udphdr_return_code = 0;
	    udp_hdr->wan_udphdr_data_len = sizeof(global_stats_t)+
				     sizeof(pipe_mgmt_stat_t);
	    break;

	case WAN_FE_SET_LB_MODE:
	case WAN_GET_MEDIA_TYPE:
	case WAN_FE_GET_STAT:
	case WAN_FE_GET_CFG:
 	case WAN_FE_FLUSH_PMON:
    	    if (IS_TE1_CARD(card)){
		sdla_te_udp(card, &udp_hdr->wan_cmd, &udp_hdr->wan_udphdr_data[0]);	
	    }else if (IS_56K_CARD(card)){
		sdla_56k_udp(card, &udp_hdr->wan_cmd, &udp_hdr->wan_udphdr_data[0]);	
	    }else{
		if (udp_hdr->wan_udphdr_command == WAN_GET_MEDIA_TYPE){
	    		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    	    		udp_hdr->wan_udphdr_data_len = sizeof(unsigned char); 
		}else{
	    		udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
		}	
	    }
	    break;
#if 0
	case WAN_FE_SET_LB_MODE:
	    /* Activate/Deactivate Line Loopback modes */
	    if (IS_TE1_CARD(card)){
		err = sdla_set_te1_lb_modes(card, 
					    udp_hdr->wan_udphdr_data[0], 
					    udp_hdr->wan_udphdr_data[1]);
	    	udp_hdr->wan_udphdr_return_code = 
				(!err) ? WAN_CMD_OK : WAN_UDP_FAILED_CMD;
	    }else{
		udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
	    }
	    udp_hdr->wan_udphdr_data_len = 0x00;
	    break;

  	case WAN_GET_MEDIA_TYPE:
	    udp_hdr->wan_udphdr_data[0] = 
		(IS_T1_CARD(card) ? WANOPT_MEDIA_T1 :
		 IS_E1_CARD(card) ? WANOPT_MEDIA_E1 :
		 IS_56K_CARD(card) ? WANOPT_MEDIA_56K : 
					WANOPT_MEDIA_NONE);
	    udp_hdr->wan_udphdr_return_code = 0;
    	    udp_hdr->wan_udphdr_data_len = sizeof(unsigned char); 
	    break;

	case WAN_FE_GET_STAT:
 	    /* TE1_56K Read T1/E1/56K alarms */
	    if (IS_TE1_CARD(card)){	
	  	*(unsigned long *)&udp_hdr->wan_udphdr_data[0] = 
						sdla_te_alarm(card, 0); 
		/* TE1 Update T1/E1 perfomance counters */
    		sdla_te_pmon(card);
	        bcopy(&card->wandev.te_pmon,
	        	  &udp_hdr->wan_udphdr_data[sizeof(unsigned long)],
			  sizeof(pmc_pmon_t));
		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	    	udp_hdr->wan_udphdr_data_len = 
			sizeof(unsigned long) + sizeof(pmc_pmon_t);

	    }else if (IS_56K_CARD(card)) {
			/* 56K Update CSU/DSU alarms */
		card->wandev.k56_alarm = sdla_56k_alarm(card, 1); 
		*(unsigned long *)&udp_hdr->wan_udphdr_data[0] = card->wandev.k56_alarm;
		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	    	udp_hdr->wan_udphdr_data_len = sizeof(unsigned long);
	    }else{
	    	udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
	    	udp_hdr->wan_udphdr_data_len = 0;
	    }
	    break;

	case WAN_FE_FLUSH_PMON:
 	    /* TE1 Flush T1/E1 pmon counters */
	    if (IS_TE1_CARD(card)){	
		sdla_flush_te1_pmon(card);
	        udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	    }else{
	    	udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
	    	udp_hdr->wan_udphdr_data_len = 0;
	    }
	    break;

	case WAN_FE_GET_CFG:
	    /* Read T1/E1 configuration */
	    if (IS_TE1_CARD(card)){	
            	bcopy(&card->wandev.te_cfg,
        	      &udp_hdr->wan_udphdr_data[0],
		      sizeof(sdla_te_cfg_t));
	    	udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    		udp_hdr->wan_udphdr_data_len = sizeof(sdla_te_cfg_t);
	    }else{
	    	udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
	    	udp_hdr->wan_udphdr_data_len = 0;
	    }
	    break;
#endif

	case WAN_GET_PROTOCOL:
	    udp_hdr->wan_udphdr_data[0] = WANCONFIG_PPP;
	    udp_hdr->wan_udphdr_return_code = 0;
	    udp_hdr->wan_udphdr_data_len = 1;
	    break;

 	case WAN_GET_PLATFORM:
	    udp_hdr->wan_udphdr_data[0] = WAN_FREEBSD_PLATFORM;
	    udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    	    udp_hdr->wan_udphdr_data_len = 1;
	    break;

udp_dflt_cmd:
	default:
	    /* it's a board command */
		WAN_MBOX_INIT(mbox);
	    mbox->wan_command = udp_hdr->wan_udphdr_command;
	    mbox->wan_data_len = udp_hdr->wan_udphdr_data_len;
	    mbox->wan_return_code = udp_hdr->wan_udphdr_return_code;
	    if (mbox->wan_data_len){ 
		memcpy(mbox->wan_data, udp_hdr->wan_udphdr_data, mbox->wan_data_len);
	    }
						    
	    /* run the command on the board */
    	    err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	    if (err != WAN_CMD_OK){ 
		ppp_error(card, err, mbox);
	        ++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_adptr_cmnd_timeout;
		break;
	    }
  	    ++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_adptr_cmnd_OK;
	    /* copy the result back to our buffer */
	    udp_hdr->wan_udphdr_command  = mbox->wan_command;
	    udp_hdr->wan_udphdr_data_len = mbox->wan_data_len;
	    udp_hdr->wan_udphdr_return_code = mbox->wan_return_code;
	    if (mbox->wan_data_len){
		memcpy(&udp_hdr->wan_udphdr_data[0], mbox->wan_data, mbox->wan_data_len);
	    }
	} /* end of switch */

	return err;
}

/*=============================================================================
 * Process UDP call of type PTPIPEAB.
 */
static int process_udp_mgmt_pkt(sdla_t *card, struct ifnet *ifp, 
				 ppp_private_area_t *ppp_priv_area) 
{
	struct mbuf*	m = NULL; 
	unsigned char 	buf2[5];
	unsigned short 	data_length = 0; 
	unsigned char*	data_addr = NULL;
	int 		udp_mgmt_req_valid = 1;
	int 		err = 0;
#ifdef ENABLE_IPV6
	int		isr = 0;
	wan_udp_pkt_t*	udp_pkt = NULL;
	struct udphdr*	udp = NULL;
	int		ip_hdr_len = 0;
	sa_family_t	af = AF_INET;
#else
	wan_udp_pkt_t*	udp_pkt = (wan_udp_pkt_t*)&ppp_priv_area->udp_pkt_data;
#endif

	bcopy(&card->wandev.udp_port, &buf2, 2);

#ifdef ENABLE_IPV6

	if ((ppp_priv_area->udp_pkt_data[0] >> 4) == 6){
		ip_hdr_len = sizeof(struct ip6_hdr);
	}else{
		ip_hdr_len = sizeof(struct ip);
	}
	udp_pkt = (wan_udp_pkt_t*)&ppp_priv_area->udp_pkt_data[ip_hdr_len];

	if (ppp_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK) {
		switch(udp_pkt->wan_udp_command) { 
		case PPIPE_GET_IBA_DATA:
		case PPP_READ_CONFIG:
		case PPP_GET_CONNECTION_INFO:
		case PPIPE_ROUTER_UP_TIME:
	    	case PPP_READ_STATISTICS:
		case PPP_READ_ERROR_STATS:
		case PPP_READ_PACKET_STATS:
		case PPP_READ_LCP_STATS:
		case PPP_READ_IPCP_STATS:
		case PPP_READ_IPXCP_STATS:
		case PPP_READ_PAP_STATS:
		case PPP_READ_CHAP_STATS:
		case PPP_READ_CODE_VERSION:
			udp_mgmt_req_valid = 1;
			break;

		default:
			udp_mgmt_req_valid = 0;
			break;
		}
	} 

	if (!udp_mgmt_req_valid){ 
		/* set length to 0 */
	  	udp_pkt->wan_udp_data_len = 0x00;
		++ ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_direction_err;
		/* set return code */
        	udp_pkt->wan_udp_return_code = 0xCD; 
	}else{ 

		err = process_udp_cmd(ifp, &udp_pkt->wan_udp_hdr);

	} /* end of else */

#else	/* without IPv6 */

	if (ppp_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK){
	       	switch(udp_pkt->wan_udp_command){ 
    		case PPIPE_GET_IBA_DATA:
	    	case PPP_READ_CONFIG:
    		case PPP_GET_CONNECTION_INFO:
	    	case PPIPE_ROUTER_UP_TIME:
    		case PPP_READ_STATISTICS:
		case PPP_READ_ERROR_STATS:
		case PPP_READ_PACKET_STATS:
		case PPP_READ_LCP_STATS:
		case PPP_READ_IPCP_STATS:
		case PPP_READ_IPXCP_STATS:
		case PPP_READ_PAP_STATS:
		case PPP_READ_CHAP_STATS:
		case PPP_READ_CODE_VERSION:
		case WAN_GET_MEDIA_TYPE:
		case WAN_FE_GET_STAT:
			udp_mgmt_req_valid = 1;
			break;

		default:
			udp_mgmt_req_valid = 0;
			break;
		}
	} 

	udp_pkt->wan_udp_opp_flag = 0;
	if (!udp_mgmt_req_valid){ 
		/* set length to 0 */
	  	udp_pkt->wan_udp_data_len = 0x00;
		++ ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_direction_err;
		/* set return code */
        	udp_pkt->wan_udp_return_code = 0xCD; 
	}else{ 

		err = process_udp_cmd(ifp, &udp_pkt->wan_udp_hdr);

	} /* end of else */

	/* Fill UDP TTL */
	udp_pkt->wan_ip_ttl = card->wandev.ttl; 
#endif /* ENABLE_IPV6 */

#ifdef ENABLE_IPV6
	data_length = reply_udp_ipv6(ppp_priv_area->udp_pkt_data, udp_pkt->wan_udp_data_len);
#else
	data_length = wan_reply_udp(card, ppp_priv_area->udp_pkt_data, udp_pkt->wan_udp_data_len);
#endif

#ifdef ENABLE_IPV6
	data_addr  = ppp_priv_area->udp_pkt_data;	

	/*
	 * Build a new mbuf chain 
	 */
	m = sdla_get(ifp, data_addr, data_length);
	if (m == NULL) {
  		++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_no_socket;
		DEBUG_EVENT("%s: no socket buffers available!\n", card->devname);
	        /* i guess nothing can be done ... the packet is gone */
		return ENOMEM;
	}
	if ((data_addr[0] >> 4) == 6){
		u_short uh_sum;
 		udp = (struct udphdr*)&data_addr[sizeof(struct ip6_hdr)];
		uh_sum = in6_cksum(m, IPPROTO_UDP, sizeof(struct ip6_hdr), 
				data_length - sizeof(struct ip6_hdr));       
		udp->uh_sum = uh_sum;
		data_addr = mtod(m, caddr_t);
 		udp = (struct udphdr*)&data_addr[sizeof(struct ip6_hdr)];
		udp->uh_sum = uh_sum;
		af = AF_INET6;
		isr = NETISR_IP6;
	}else{
		af = AF_INET;
		isr = NETISR_IP;
	}
#endif

	wan_bpf_report(ifp, m, 1);
	if (ppp_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK){ 
#ifdef ENABLE_IPV6
		wan_skb_free(m);	/*m_freem(m);*/
#endif
		++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_passed_to_adptr;
		ppp_send(card, ppp_priv_area->udp_pkt_data, data_length, ppp_priv_area->protocol);
	}else{	/* Pass it up the stack */
#if defined(__OpenBSD__) || (__FreeBSD_version < 501000)
		wan_spinlock_t	s;
#endif
		/*
		 * Ok, what we've got now in scp->umg_top_buff is a fully formatted IP 
		 * packet which consists of responce to the UDP management request. The 
		 * length of the packet is 'data_length'. We have to free the original 
		 * mbuf and allocate the new one. ( when i tried to re-use the same 
		 * mbuf with m_pullup i ended up crashing the system somewhere in IP
		 * stack after i passed the frame up - looks like they free the buffer
		 * anyways if the driver does not ).
		 */  
#ifdef ENABLE_IPV6

# if (__FreeBSD_version >= 501000)
		netisr_queue(isr, m);
# else
		if ((err = family_enqueue(af, m))){
			DEBUG_EVENT("%s: Failed to send packet to IP stack (%d)!\n", 
					card->devname, err); 
            		return -EINVAL;
		}
# endif
#else
	        data_addr  = ppp_priv_area->udp_pkt_data;	

        	/*
	         * Build a new mbuf chain 
        	 */
		m = sdla_get(ifp, data_addr, data_length);
		if (m == NULL) {
			++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_no_socket;
			DEBUG_EVENT("%s: no socket buffers available!\n", card->devname);
			/* i guess nothing can be done ... the packet is gone */
			return ENOMEM;
		}
        
		/* 
		 * Ok, at this point we've got an mbuf chain pointed by m which contains
	         * a UDP management reply - now we just have to pass it up the stack -
        	 * and that's it !!!
	         * Note that we neither report those packets to BPF nore update any 
        	 * stats. Those packets are meant to be invisible ;-)
	         * Also we run inside splimp already - do there's no need to call it 
        	 * again.
		 */
# if (__FreeBSD_version >= 501000)
		netisr_queue(NETISR_IP, m);
# else
		wan_spin_lock_irq(NULL, &s);
	        schednetisr(NETISR_IP);
        	if (IF_QFULL(&ipintrq)){
			/* oh, no - IP queue is full - well - screw it. We cant do much 
			 * here. 
			 */
			DEBUG_RX("%s: IP queue is full, drop the packet!\n", 
					card->devname);
			IF_DROP(&ipintrq);
			wan_skb_free(m);	/*m_freem(m);*/
			wan_spin_unlock_irq(NULL, &s);
			return -EINVAL;
		}

		/* ok enqueue the packet */
		IF_ENQUEUE(&ipintrq, m);
		wan_spin_unlock_irq(NULL, &s);
# endif
#endif

		DEBUG_UDP("%s: UDP packet passed back on IP's queue\n", 
						card->devname);
		++ppp_priv_area->pipe_mgmt_stat.UDP_PIPE_mgmt_passed_to_stack;
	}

	ppp_priv_area->udp_pkt_lgth = 0;
	return 0; 
}

/*=============================================================================
 * Initial the ppp_private_area structure.
 */
static void init_ppp_priv_struct( ppp_private_area_t *ppp_priv_area )
{
	bzero(&ppp_priv_area->if_send_stat, sizeof(if_send_stat_t));
	bzero(&ppp_priv_area->rx_intr_stat, sizeof(rx_intr_stat_t));
	bzero(&ppp_priv_area->pipe_mgmt_stat, sizeof(pipe_mgmt_stat_t));
}

/*============================================================================
 * Initialize Global Statistics
 */
static void init_global_statistics( sdla_t *card )
{
	bzero(&card->statistics, sizeof(global_stats_t));
}

/*============================================================================
 * Initialize Receive and Transmit Buffers.
 */
static void init_ppp_tx_rx_buff( sdla_t *card )
{
	ppp508_buf_info_t	info;

	card->hw_iface.peek(card->hw, PPP514_BUF_OFFS, &info, sizeof(ppp508_buf_info_t));
        card->u.p.txbuf_base_off = info.txb_ptr;
       	card->u.p.txbuf_last_off = 
			card->u.p.txbuf_base_off + (info.txb_num - 1) * sizeof(ppp_buf_ctl_t); 

        card->u.p.rxbuf_base_off = info.rxb_ptr;
        card->u.p.rxbuf_last_off = 
			card->u.p.rxbuf_base_off + (info.rxb_num - 1) * sizeof(ppp_buf_ctl_t);

	card->u.p.rx_base_off = info.rxb_base;
	card->u.p.rx_top_off  = info.rxb_end;
	card->u.p.txbuf_off = card->u.p.txbuf_base_off;
	card->rxmb_off = card->u.p.rxbuf_base_off;
}

/*=============================================================================
 * Read Connection Information (ie for Remote IP address assginment).
 * Called when ppp interface connected.
 */
static int read_info(sdla_t *card)
{
	netdevice_t		*ifp;
	struct ifaddr		*ifa = NULL; 
	struct sockaddr_in	*saddr = NULL;
	ppp_private_area_t	*ppp_priv_area = NULL;
	int 			err = 0;

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	WAN_ASSERT(ifp == NULL);
	WAN_ASSERT(wan_netif_priv(ifp) == NULL);
	ppp_priv_area = wan_netif_priv(ifp);

	ifa = WAN_TAILQ_FIRST(ifp); 
	if ((ifa = WAN_TAILQ_NEXT(ifa)) != NULL){
		/* Local IP address */
		saddr = (struct sockaddr_in*)ifa->ifa_addr;
		if (saddr){
			/* write new local IP address */
			saddr->sin_addr.s_addr = ppp_priv_area->ip_local;
		}
		/* Remote IP address */
		saddr = (struct sockaddr_in*)ifa->ifa_dstaddr;
		if (saddr){
			/* write new remote IP address */
			saddr->sin_addr.s_addr = ppp_priv_area->ip_local;
		}
		/* Add a new route instead od deleted one */
		err = rtrequest((int)RTM_ADD,			/* action      */
			  ifa->ifa_dstaddr,			/* destination */
			  ifa->ifa_addr,			/* gateway     */
			  ifa->ifa_netmask,			/* mask        */
			  RTF_HOST,				/* flag	       */ 	
	        	  (struct rtentry**)0);
	}

	if (!err){
		DEBUG_EVENT("%s: Assigned Lcl. Addr: %s\n", 
				card->devname, in_ntoa(ppp_priv_area->ip_local));
		DEBUG_EVENT("%s: Assigned Rmt. Addr: %s\n", 
				card->devname, in_ntoa(ppp_priv_area->ip_local));
	}else{ 
		DEBUG_EVENT("%s: Adding of route failed:\n", card->devname);
		DEBUG_EVENT("%s:	Local : %s\n", 
				card->devname,in_ntoa(ppp_priv_area->ip_local));
		DEBUG_EVENT("%s:	Remote: %s\n",
				card->devname,in_ntoa(ppp_priv_area->ip_remote));
		DEBUG_EVENT("%s:   An error occurred in IP assignment.\n", 
							card->devname);
	}
	return 0;
}


/*=============================================================================
 * Remove Dynamic Route.
 * Called when ppp interface disconnected.
 */
static int remove_route( sdla_t *card )
{
	/* 
	 * ok, i was too lazy to allocate new sockaddr structures - so i'm 
	 * going to temporary replace interface ip addresses, then call delete 
	 * route and then restore them back.
	 * Note that all this is done only if old and new addresses are different.
	 */
	netdevice_t		*ifp;
	ppp_private_area_t	*ppp_priv_area = NULL;
	struct ifaddr		*ifa = NULL;
	int 			err = 0;

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	WAN_ASSERT(ifp == NULL);
	WAN_ASSERT(wan_netif_priv(ifp) == NULL);
	ppp_priv_area = wan_netif_priv(ifp);

	/*  get the next one - its where all the stuff is */
	ifa = WAN_TAILQ_FIRST(ifp);
	if ((ifa = WAN_TAILQ_NEXT(ifa)) != NULL){
        	/* if the entry is not found. we'll give it another try then 	  */
	        if ((err = rtrequest((int)RTM_DELETE, 		/* action */
				 ifa->ifa_dstaddr,		/* destination */
				 ifa->ifa_addr,		/* gateway */
				 ifa->ifa_netmask,		/* mask */
				 RTF_HOST,			/* flag */
				 (struct rtentry **)0)) == 0)
		{
			DEBUG_EVENT("%s: PPP deleting dynamic route successfuly!\n",
					card->devname); 
	        }else{
			DEBUG_EVENT("%s: Deleting dynamic route failed %d!\n",
					card->devname, err);
		}
	}
	return err;
}


#ifdef PPP_OLD_INTR_TEST
/*
 * ============================================================================
 * Perform the Interrupt Test by running the READ_CODE_VERSION command MAX_INTR
 * _TEST_COUNTER times.
 */
static int ppp_intr_test( sdla_t *card )
{
	int 		err, i;

	Intr_test_counter = 0;
 
	err = ppp_set_intr_mode(card, PPP_INTR_CMD);
	if (err == WAN_CMD_OK) { 
    		wan_mbox_t*	mbox = &card->wan_mbox;
		WAN_MBOX_INIT(mbox);
		for (i = 0; i < MAX_INTR_TEST_COUNTER; i ++) { 
			/* Run command READ_CODE_VERSION */
			mbox->wan_data_len  = 0;
			mbox->wan_command = PPP_READ_CODE_VERSION;
    			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
			if (err != WAN_CMD_OK) 
				ppp_error(card, err, mbox);
		}
		err = ppp_set_intr_mode(card, 0x0);
	}

	if (err || (Intr_test_counter < MAX_INTR_TEST_COUNTER)){
		DEBUG_EVENT("%s: Interrupt Test Failed, Counter: %d\n",
					card->devname, Intr_test_counter);
		DEBUG_EVENT("%s: Please choose another interrupt!!\n", 
					card->devname); 
		return -EIO;
	}
	DEBUG_EVENT("%s: Interrupt Test Passed, Counter: %d\n",
					card->devname, Intr_test_counter);
 
	return 0;
}
#endif

/*==============================================================================
 * Determine what type of UDP call it is. DRVSTATS or PTPIPEAB ?
 */
static int udp_pkt_type(sdla_t* card, struct mbuf *m0)
{
    unsigned char*	sendpacket = NULL;
    unsigned char 	buf2[5]; 
    /* FIXME: Use the structure */
#ifdef ENABLE_IPV6
    struct udphdr*	udp = NULL;
    wan_mgmt_t*		mgmt = NULL;
    wan_cmd_t*		cmd = NULL;
#else
    wan_udp_pkt_t*	udp_pkt = mtod(m0, wan_udp_pkt_t *); 
#endif

    if (wan_skb_len(m0) < sizeof(wan_udp_pkt_t)){
    	return UDP_INVALID_TYPE;
    }
    sendpacket = mtod(m0, char*);
    bcopy(&card->wandev.udp_port, &buf2, 2);

#ifdef ENABLE_IPV6
    if ((mtod(m0, caddr_t)[0] >> 4) == 6 || (mtod(m0, caddr_t)[0] >> 4) == 4){
	if ((mtod(m0, caddr_t)[0] >> 4) == 6){
		struct ip6_hdr* ip6 = (struct ip6_hdr*)mtod(m0, caddr_t);
		if (ip6->ip6_nxt != UDPMGMT_UDP_PROTOCOL) 
		    return UDP_INVALID_TYPE;
		udp = (struct udphdr*)&(mtod(m0, caddr_t)[sizeof(struct ip6_hdr)]);
		mgmt = (wan_mgmt_t*)&(mtod(m0, caddr_t)[sizeof(struct ip6_hdr)+sizeof(struct udphdr)]);
		cmd = (wan_cmd_t*)&(mtod(m0, caddr_t)[sizeof(struct ip6_hdr)+sizeof(struct udphdr)+sizeof(struct wan_mgmt_t)]);
  	}else{
		struct ip* ip4 = (struct ip*)mtod(m0, caddr_t);
		if (ip4->ip_p != UDPMGMT_UDP_PROTOCOL) 
		    return UDP_INVALID_TYPE;
		udp = (struct udphdr*)&(mtod(m0, caddr_t)[sizeof(struct ip)]);
		mgmt = (wan_mgmt_t*)&(mtod(m0, caddr_t)[sizeof(struct ip)+sizeof(struct udphdr)]);
		cmd = (wan_mgmt_t*)&(mtod(m0, caddr_t)[sizeof(struct ip)+sizeof(struct udphdr)+sizeof(struct wan_mgmt_t)]);
	}
	if (mgmt->wan_udp_request_reply == UDPMGMT_REQUEST){
        	if ((!strncmp(mgmt->wan_udp_signature, PPP_UDP_SIGNATURE, PPP_UDP_SIGNATURE_LEN) ||
		    !strncmp(mgmt->wan_udp_signature, GLOBAL_UDP_SIGNATURE, GLOBAL_UDP_SIGNATURE_LEN)) &&
		    (udp->wan_udp_dport == ntohs(card->wandev.udp_port)))
		{
			return UDP_PTPIPE_TYPE;
		}
        	if (!strncmp(mgmt->wan_udp_signature, UD_DRVRSTATS_SIGNATURE, PPP_UDP_SIGNATURE_LEN) &&
		    (udp->wan_udp_dport == ntohs(card->wandev.udp_port)))
		{
			return UDP_DRVSTATS_TYPE;
		}
	        if (!strncmp(mgmt->wan_udp_signature, GLOBAL_UDP_SIGNATURE, GLOBAL_UDP_SIGNATURE_LEN) && 
		    cmd->wan_cmd_command == WAN_GET_PROTOCOL)
		{	
		    return UDP_PTPIPE_TYPE;
		}
	}
    }
    return UDP_INVALID_TYPE;

#else

    if (udp_pkt->wan_ip_v  == 0x4 &&     /* IP packet */ 
	udp_pkt->wan_ip_p  == UDPMGMT_UDP_PROTOCOL &&        /* UDP packet */
        udp_pkt->wan_udp_request_reply == UDPMGMT_REQUEST)
    {

        if ((!strncmp(udp_pkt->wan_udp_signature, PPP_UDP_SIGNATURE, PPP_UDP_SIGNATURE_LEN) ||
	    !strncmp(udp_pkt->wan_udp_signature, GLOBAL_UDP_SIGNATURE, GLOBAL_UDP_SIGNATURE_LEN)) &&
	    (udp_pkt->wan_udp_dport == ntohs(card->wandev.udp_port)))
	{
		return UDP_PTPIPE_TYPE;
	}
        if (!strncmp(udp_pkt->wan_udp_signature, UDP_DRVRSTATS_SIGNATURE, PPP_UDP_SIGNATURE_LEN) &&
	    (udp_pkt->wan_udp_dport == ntohs(card->wandev.udp_port)))
	{
		return UDP_DRVSTATS_TYPE;
	}
        if (!strncmp(udp_pkt->wan_udp_signature, GLOBAL_UDP_SIGNATURE, GLOBAL_UDP_SIGNATURE_LEN) && 
	    udp_pkt->wan_udp_command == WAN_GET_PROTOCOL)
	{
	    return UDP_PTPIPE_TYPE;
	}
    } 

    return UDP_INVALID_TYPE;

#endif

}

/*============================================================================
 * Check to see if the packet to be transmitted contains a broadcast or
 * multicast source IP address.
 */
static int chk_bcast_mcast_addr(sdla_t *card, struct ifnet* ifp, struct mbuf *m)
{
	uint32_t src_ip_addr = 0;
	uint32_t broadcast_ip_addr = 0;
	struct ifaddr *ifa = WAN_TAILQ_FIRST(ifp);
	struct sockaddr_in *saddr = NULL;

	/* read the IP source address from the outgoing packet */
	src_ip_addr = *(uint32_t *)(mtod(m, caddr_t) + 12);

	/* read the IP broadcast address for the device */
	if ((saddr = (struct sockaddr_in *)ifa->ifa_broadaddr) != NULL) 
		broadcast_ip_addr = htonl(saddr->sin_addr.s_addr);

	/* check if the IP Source Address is a Broadcast address */
	if((ifp->if_flags & IFF_BROADCAST) && (src_ip_addr == broadcast_ip_addr)){ 
		DEBUG_EVENT("%s: Broadcast Source Address silently discarded\n",
							card->devname);
		return 1;
	} 

	/* check if the IP Source Address is a Multicast address */
	if ((ntohl(src_ip_addr) >= 0xE0000001) && 
		 (ntohl(src_ip_addr) <= 0xFFFFFFFE))
	{
		DEBUG_EVENT("%s: Multicast Source Address silently discarded\n", 
					card->devname);
		return 1;
	}

	return 0;
}

/****** PPP Firmware Interface Functions ************************************/

/*============================================================================
 * Read TE1/56K Front end registers
 */
static unsigned char 
read_front_end_reg (void* card_id, unsigned short reg)
{
	sdla_t* 	card = (sdla_t*)card_id;
    	wan_mbox_t*	mbox = &card->wan_mbox;
	char*		data = mbox->wan_data;
	int 		err = 0x0;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(FRONT_END_REG_STRUCT);
	mbox->wan_command = READ_FRONT_END_REGISTER;
	((FRONT_END_REG_STRUCT *)data)->register_number = (unsigned short)reg;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		ppp_error(card, err, mbox);
	}
	return(((FRONT_END_REG_STRUCT *)data)->register_value);
}

/*============================================================================
 * Write to TE1/56K Front end registers 
 */
static unsigned char 
write_front_end_reg (void* card_id, unsigned short reg, unsigned char value)
{
	sdla_t* 	card = (sdla_t*)card_id;
    	wan_mbox_t*	mbox = &card->wan_mbox;
	char*		data = mbox->wan_data;
	int 			err;
	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(FRONT_END_REG_STRUCT);
	mbox->wan_command = WRITE_FRONT_END_REGISTER;
	((FRONT_END_REG_STRUCT *)data)->register_number = (unsigned short)reg;
	((FRONT_END_REG_STRUCT *)data)->register_value = value;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		ppp_error(card, err, mbox);
	}
	return err;
}

/*============================================================================
 * TE1 PPP Enable hardware timer interrupt. 
 */
static void ppp_enable_timer(void* card_id)
{
	netdevice_t		*ifp;
	sdla_t	 		*card = (sdla_t*)card_id;
	ppp_private_area_t	*ppp_priv_area = NULL;
	
	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	WAN_ASSERT1(ifp == NULL);
	WAN_ASSERT1(wan_netif_priv(ifp) == NULL);
	ppp_priv_area = wan_netif_priv(ifp);
	ppp_priv_area->timer_int_enabled |= TMR_INT_ENABLED_TE;
	card->hw_iface.set_bit(card->hw, card->intr_perm_off, PPP_INTR_TIMER);
	return;
}

static int set_adapter_config (sdla_t* card)
{
    	wan_mbox_t*			mbox = &card->wan_mbox;
	ADAPTER_CONFIGURATION_STRUCT*	cfg = (ADAPTER_CONFIGURATION_STRUCT*)mbox->wan_data; 
	int				err;
	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(ADAPTER_CONFIGURATION_STRUCT);
	mbox->wan_command = SET_ADAPTER_CONFIGURATION;
	card->hw_iface.getcfg(card->hw, SDLA_ADAPTERTYPE, &cfg->adapter_type); 
	cfg->adapter_config = 0x00; 
	cfg->operating_frequency = 0x00; 
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK)
		ppp_error(card, err, mbox);

	return err;
}


/*============================================================================
 * Read firmware code version.
 *	Put code version as ASCII string in str. 
 */
static int ppp_read_version(sdla_t *card, char *str)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_command = PPP_READ_CODE_VERSION;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		ppp_error(card, err, mbox);
	}else{ 
		if (str){ 
			memcpy(str, mbox->wan_data, mbox->wan_data_len);
			str[mbox->wan_data_len] = '\0';
		}
	}

	return err;
}

#ifdef ENABLE_IPV6

/*===========================================================================
 * Set Out-Bound Authentication.
*/
static int ppp_set_outbnd_chap_auth (sdla_t *card, ppp_private_area_t *ppp_priv_area)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;

	WAN_MBOX_INIT(mbox);
	bzero(mbox->wan_data, (strlen(ppp_priv_area->sysname) +  
			strlen(ppp_priv_area->chap_userid) + 
			strlen(ppp_priv_area->chap_passwd) + 3));
	memcpy(mbox->wan_data, ppp_priv_area->sysname, strlen(ppp_priv_area->sysname));
	memcpu(mbox->wan_data + strlen(ppp_priv_area->sysname) + 1,
			ppp_priv_area->chap_userid, strlen(ppp_priv_area->chap_userid));
	memcpy(mbox->wan_data + strlen(ppp_priv_area->sysname) + strlen(ppp_priv_area->chap_userid + 2),
  			ppp_priv_area->chap_passwd, strlen(ppp_priv_area->chap_passwd));
	mbox->wan_data_len = strlen(ppp_priv_area->sysname) + 
				strlen(ppp_priv_area->chap_userid) + 
				strlen(ppp_priv_area->chap_passwd) + 3;
	mbox->wan_command = PPP_SET_OUTBOUND_AUTH;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		ppp_error(card, err, mbox);
	}
	return err;
}


/*===========================================================================
 * Set In-Bound Authentication.
*/
static int ppp_set_inbnd_chap_auth (sdla_t *card, ppp_private_area_t *ppp_priv_area)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err, i;
	char* 	user_tokens[32];
	char* 	pass_tokens[32];
	int 	userids, passwds;
	int 	add_ptr;

	WAN_MBOX_INIT(mbox);
	bzero(mbox->wan_data, 1008);
	memcpy(mbox->wan_data, ppp_priv_area->sysname, strlen(ppp_priv_area->sysname));
	
	/* Parse the userid string and the password string and build a string
	 * to copy it to the data area of the command structure.   The string
	 * will look like "SYS_NAME<NULL>USER1<NULL>PASS1<NULL>USER2<NULL>PASS2
	 * ....<NULL> " 
	 */
	userids = tokenize( ppp_priv_area->chap_userid, user_tokens);
	passwds = tokenize( ppp_priv_area->chap_passwd, pass_tokens);

	if (userids != passwds){
     		DEBUG_EVENT("%s: Number of passwords does not equal the number of user ids\n",
							 card->devname);
		return 1;
	}

	add_ptr = strlen(ppp_priv_area->sysname) + 1;
	for (i=0; i<userids; i++){
		memcpy(mbox->wan_data +_ add_ptr, user_tokens[i], strlen(user_tokens[i]));
		memcpy(mbox->wan_data + offsetof(wan_mbox_t, wan_data) + add_ptr + strlen(user_tokens[i])+1, 
				pass_tokens[i], strlen(pass_tokens[i]));
		add_ptr = add_ptr + strlen(user_tokens[i]) + 1 + 
				strlen(pass_tokens[i]) + 1;
	}

	mbox->wan_data_len  = add_ptr + 1;
	mbox->wan_command = PPP_SET_INBOUND_AUTH;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK)
		ppp_error(card, err, mbox);

	return err;
}

/*===========================================================================
 * Set Out-Bound Authentication.
*/
static int ppp_set_outbnd_pap_auth (sdla_t *card, ppp_private_area_t *ppp_priv_area)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int err;

	WAN_MBOX_INIT(mbox);
	bzero(mbox->wan_data, (strlen(ppp_priv_area->sysname) +  
			strlen(ppp_priv_area->pap_userid) + 
			strlen(ppp_priv_area->pap_passwd) + 3));
	memcpy(mbox->wan_data, ppp_priv_area->sysname, strlen(ppp_priv_area->sysname));
	memcpy(mbox->wan_data + strlen(ppp_priv_area->sysname) + 1,
			ppp_priv_area->pap_userid, strlen(ppp_priv_area->pap_userid));
	memcpy(mbox->wan_data + strlen(ppp_priv_area->sysname) +
			strlen(ppp_priv_area->pap_userid + 2),
  			ppp_priv_area->pap_passwd, strlen(ppp_priv_area->pap_passwd));
	mbox->wan_data_len = strlen(ppp_priv_area->sysname) + 
			strlen(ppp_priv_area->pap_userid) + 
			strlen(ppp_priv_area->pap_passwd) + 3;
	mbox->wan_command = PPP_SET_OUTBOUND_PAP_AUTH;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK)
		ppp_error(card, err, mbox);

	return err;
}


/*===========================================================================
 * Set In-Bound Authentication.
*/
static int ppp_set_inbnd_pap_auth (sdla_t *card, ppp_private_area_t *ppp_priv_area)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err, i;
	char* 	user_tokens[32];
	char* 	pass_tokens[32];
	int 	userids, passwds;
	int 	add_ptr;

	WAN_MBOX_INIT(mbox);
	bzero(mbox->wan_data, 1008);
	memcpy(mbox->wan_data, ppp_priv_area->sysname, strlen(ppp_priv_area->sysname));
	
	/* Parse the userid string and the password string and build a string
	 * to copy it to the data area of the command structure.   The string
	 * will look like "SYS_NAME<NULL>USER1<NULL>PASS1<NULL>USER2<NULL>PASS2
	 * ....<NULL> " 
	 */
	userids = tokenize( ppp_priv_area->pap_userid, user_tokens);
	passwds = tokenize( ppp_priv_area->pap_passwd, pass_tokens);

	if (userids != passwds){
        	DEBUG_EVENT("%s: Number of passwords does not equal the number of user ids\n",
							 card->devname);
		return 1;
	}

	add_ptr = strlen(ppp_priv_area->sysname) + 1;
	for (i=0; i<userids; i++){
		memcpy(mbox->wan_data + add_ptr, user_tokens[i], strlen(user_tokens[i]));
		memcpy(mbox->wan_data + add_ptr + strlen(user_tokens[i]) + 1, 
				pass_tokens[i], strlen(pass_tokens[i]));
		add_ptr = add_ptr + strlen(user_tokens[i]) + 1 + 
				strlen(pass_tokens[i]) + 1;
	}

	mbox->wan_data_len  = add_ptr + 1;
	mbox->wan_command = PPP_SET_INBOUND_PAP_AUTH;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK)
		ppp_error(card, err, mbox);

	return err;
}


#else

/*===========================================================================
 * Set Out-Bound Authentication.
*/
static int ppp_set_outbnd_auth (sdla_t *card, ppp_private_area_t *ppp_priv_area)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int err;

	WAN_MBOX_INIT(mbox);
	bzero(mbox->wan_data, (strlen(ppp_priv_area->sysname) +
			strlen(ppp_priv_area->userid) + 
			strlen(ppp_priv_area->passwd) + 3));
	memcpy(mbox->wan_data, ppp_priv_area->sysname, strlen(ppp_priv_area->sysname));
	memcpy(mbox->wan_data + strlen(ppp_priv_area->sysname) + 1,
			ppp_priv_area->userid, strlen(ppp_priv_area->userid));
	memcpy(mbox->wan_data + strlen(ppp_priv_area->sysname) +
			strlen(ppp_priv_area->userid + 2),
  			ppp_priv_area->passwd, strlen(ppp_priv_area->passwd));
	mbox->wan_data_len = strlen(ppp_priv_area->sysname) + 
			strlen(ppp_priv_area->userid) + 
			strlen(ppp_priv_area->passwd) + 3;
	mbox->wan_command = PPP_SET_OUTBOUND_AUTH;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK)
		ppp_error(card, err, mbox);

	return err;
}


/*===========================================================================
 * Set In-Bound Authentication.
*/
static int ppp_set_inbnd_auth (sdla_t *card, ppp_private_area_t *ppp_priv_area)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err, i;
	char* 	user_tokens[32];
	char* 	pass_tokens[32];
	int 	userids, passwds;
	int 	add_ptr;

	WAN_MBOX_INIT(mbox);
	bzero(mbox->wan_data, 1008);
	memcpy(mbox->wan_data, ppp_priv_area->sysname, strlen(ppp_priv_area->sysname));
	
	/* Parse the userid string and the password string and build a string
	 * to copy it to the data area of the command structure.   The string
	 * will look like "SYS_NAME<NULL>USER1<NULL>PASS1<NULL>USER2<NULL>PASS2
	 * ....<NULL> " 
	 */
	userids = tokenize( ppp_priv_area->userid, user_tokens);
	passwds = tokenize( ppp_priv_area->passwd, pass_tokens);

	if (userids != passwds){
        	DEBUG_EVENT("%s: Number of passwords does not equal the number of user ids\n",
							 card->devname);
		return 1;
	}

	add_ptr = strlen(ppp_priv_area->sysname) + 1;
	for (i=0; i<userids; i++){
		memcpy(mbox->wan_data + add_ptr, user_tokens[i], strlen(user_tokens[i]));
		memcpy(mbox->wan_data + add_ptr + strlen(user_tokens[i]) + 1, 
				pass_tokens[i], strlen(pass_tokens[i]));
		add_ptr = add_ptr + strlen(user_tokens[i]) + 1 + 
				strlen(pass_tokens[i]) + 1;
	}

	mbox->wan_data_len  = add_ptr + 1;
	mbox->wan_command = PPP_SET_INBOUND_AUTH;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK)
		ppp_error(card, err, mbox);

	return err;
}
#endif

#if 0
static char *my_strtok(char *p, const char *tok)
{
    static char *phead = NULL, *tmp = NULL;
    const char *tk = NULL;

    if (p)
    {
	if (phead)
	    wan_free(phead);
	phead = wan_malloc(strlen(p)+1);
	if (phead == NULL)
	    return NULL;
	bcopy(p, phead, strlen(p));
	tmp = phead;
    }
    for(;; ++tmp)
    {
	tk = tok;
	for(;*tk != '\0'; ++tk)
	    if (*tmp == *tk)
		return tmp;
	if (!*tmp)
	    return NULL; 
    }
	
}
#endif
/*
 * ============================================================================
 * Tokenize string.
 *      Parse a string of the following syntax:
 *              <arg1>,<arg2>,...
 *      and fill array of tokens with pointers to string elements.
 *
 */
static int tokenize (char *str, char **tokens)
{
	int cnt = 0;
	char* eos = NULL;

	if (str == NULL)
		return 0;
	tokens[0] = str;
	if (*tokens[cnt] == '/')
		tokens[cnt]++;
	while(tokens[cnt] && (cnt < 32 - 1)){
		eos = my_strchr(tokens[cnt], '/');
		tokens[++cnt] = eos;
		if (eos){
			tokens[cnt] = eos + 1;
			*eos = '\0';
		}
	}
#if 0
    tokens[0] = my_strtok(str, "/");
    while (tokens[cnt] && (cnt < 32 - 1))
    {
        tokens[cnt] = strstrip(tokens[cnt], " \t");
        tokens[++cnt] = my_strtok(NULL, "/");
    }
#endif
	return cnt;
}

/*
 * ============================================================================
 *
 */
static char *my_strchr(char *p, char ch)
{
	for(;; ++p){
		if (*p == ch)
			return ((char*)p);
		if (!(*p))
			return NULL;
	}
	return NULL;
}

/*============================================================================
 * Strip leading and trailing spaces off the string str.
 */
#if 0
static char* strstrip (char *str, char* s)
{
    char *eos = str + strlen(str);          /* -> end of string */

    while (*str && my_strchr(s, *str))
        ++str;                           /* strip leading spaces */
    
    while ((eos > str) && my_strchr(s, *(eos - 1)))
        --eos;                           /* strip trailing spaces */
     
    *eos = '\0';
    return str;
}
#endif
/*============================================================================
 * Configure PPP firmware.
 */
static int ppp_configure(sdla_t *card, void *data)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	data_len = sizeof(ppp508_conf_t); 
	int 	err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len  = data_len;
	mbox->wan_command = PPP_SET_CONFIG;
	memcpy(mbox->wan_data, data, data_len);
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	if (err != WAN_CMD_OK) 
		ppp_error(card, err, mbox);
	
	return err;
}

/*============================================================================
 * Set interrupt mode.
 */
static int ppp_set_intr_mode(sdla_t *card, unsigned char mode)
{
    	wan_mbox_t*		mbox = &card->wan_mbox;
	ppp_intr_info_t*	ppp_intr_data = (ppp_intr_info_t*)mbox->wan_data;
	int 			err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 2;
	mbox->wan_command = PPP_SET_INTR_FLAGS;

	ppp_intr_data->irq = card->wandev.irq;
	ppp_intr_data->i_enable = mode;
	/* If timer has been enabled, set the timer delay to 1sec */
	if (mode & PPP_INTR_TIMER){
	   	ppp_intr_data->timer_len = 5;/* 100; 250; */
	        mbox->wan_data_len = 4;
	}
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	if (err != WAN_CMD_OK) 
		ppp_error(card, err, mbox);
 
	return err;
}

/*
 * ===========================================================================
 * Read interrupt mode
 *
 */
#if 0
static int ppp_read_intr_mode(sdla_t* card)
{
	ppp_mbox_t*	mbox = &card->wan_mbox;
	int 	err = 0;
 
	mbox->wan_command = PPP_READ_INTR_FLAGS;
	mbox->wan_data_len  = 0x0;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){ 
		ppp_error(card, err, mbox);
	}else{
        	ppp_intr_info_t *ppp_intr_data = (ppp_intr_info_t *) &mbox->wan_data[0];
		return ppp_intr_data->i_enable;	
	}
	return err;
}
#endif
 
/*============================================================================
 * Enable communications.
 */
static int ppp_comm_enable(sdla_t *card)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_command = PPP_COMM_ENABLE;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		if (err == PPPRES_COMMS_ENABLED){
			DEBUG_CMD("%s: Communication already enabled\n", 
						card->devname);
		}else{
			ppp_error(card, err, mbox);
		}
	}else{
		DEBUG_CMD("%s: Communication enabled.\n", card->devname);
	}
	
	return err;
}

/*============================================================================
 * Disable communications.
 */
static int ppp_comm_disable(sdla_t *card)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_command = PPP_COMM_DISABLE;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){ 
		ppp_error(card, err, mbox);
	}else{
		DEBUG_CMD("%s: Communication disabled.\n", card->devname);
	}

	return err;
}

/*============================================================================
 * Get communications error statistics.
 */
static int ppp_get_err_stats(sdla_t *card, ppp_err_stats_t* stats)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_command = PPP_READ_ERROR_STATS;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err == WAN_CMD_OK){
		memcpy(stats, mbox->wan_data, sizeof(ppp_err_stats_t));
	}else{
		ppp_error(card, err, mbox);
	}
	
	return err;
}

/*============================================================================
 * Get communications error statistics.
 */
static int ppp_get_stats(sdla_t *card, ppp508_stats_t* stats)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_command = PPP_READ_STATISTICS;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	
	if (err == WAN_CMD_OK){ 
		memcpy(stats, mbox->wan_data, sizeof(ppp508_stats_t));
	}else{
		ppp_error(card, err, mbox);
	} 
	
	return err;
}

static int read_connection_info (sdla_t *card)
{
	netdevice_t		*ifp;
	ppp_private_area_t	*ppp_priv_area = NULL;
    	wan_mbox_t		*mbox = &card->wan_mbox;
	ppp508_connect_info_t	ppp508_connect_info;
	int 			err;

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	WAN_ASSERT(ifp == NULL);
	WAN_ASSERT(wan_netif_priv(ifp) == NULL);
	ppp_priv_area = wan_netif_priv(ifp);
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len  = 0;
	mbox->wan_command = PPP_GET_CONNECTION_INFO;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		ppp_error(card, err, mbox);
	}else{
		memcpy(&ppp508_connect_info, mbox->wan_data, sizeof(ppp508_connect_info_t));
		ppp_priv_area->ip_remote = ppp508_connect_info.ip_remote;
		ppp_priv_area->ip_local = ppp508_connect_info.ip_local;
	}

	return err;
}

static int ppp_get_pkt_stats(sdla_t* card, ppp_pkt_stats_t* pkt_stats)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
    	int err;
	
	WAN_MBOX_INIT(mbox);
    	mbox->wan_command = PPP_READ_PACKET_STATS;
    	mbox->wan_data_len  = 0;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
    	if (err != WAN_CMD_OK){
		ppp_error(card, err, mbox);
    	}else{
		if (pkt_stats){
			memcpy(pkt_stats, mbox->wan_data, sizeof(ppp_pkt_stats_t));
		}
	}
    	return err;
}

static int ppp_get_lcp_stats(sdla_t* card, ppp_lcp_stats_t* lcp_stats)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
    	int err;
	
	WAN_MBOX_INIT(mbox);
    	mbox->wan_command = PPP_READ_LCP_STATS;
    	mbox->wan_data_len  = 0;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
    	if (err != WAN_CMD_OK){
		ppp_error(card, err, mbox);
    	}else{
		if (lcp_stats){
			memcpy(&lcp_stats, mbox->wan_data, sizeof(ppp_lcp_stats_t));
		}
	}
    	return err;
}

/*============================================================================
 * Send packet.
 *	Return:	0 - o.k.
 *		1 - no transmit buffers available
 */
static int ppp_send (sdla_t *card, void *data, unsigned len, unsigned proto)
{
    	ppp_buf_ctl_t	txbuf;

    	card->hw_iface.peek(card->hw, card->u.p.txbuf_off, &txbuf, sizeof(ppp_buf_ctl_t));
    	if (txbuf.flag){
        	return 1;
	}
    	card->hw_iface.poke(card->hw, txbuf.buf.ptr, data, len);
    	if (proto == htons(ETHERTYPE_IPX)){
		txbuf.proto = 0x01;	/* protocol ID */
    	}else{
		txbuf.proto = 0x00;	/* protocol ID */
    	}
    	txbuf.length = len;
    	card->hw_iface.poke(card->hw, card->u.p.txbuf_off, &txbuf, sizeof(txbuf)); 
    	/* start transmission */
    	txbuf.flag = 0x01;
    	card->hw_iface.poke(card->hw, card->u.p.txbuf_off, &txbuf, sizeof(txbuf)); 

    	/* Update transmit buffer control fields */
    	card->u.p.txbuf_off += sizeof(ppp_buf_ctl_t);
    	if (card->u.p.txbuf_off > card->u.p.txbuf_last_off){
		card->u.p.txbuf_off = card->u.p.txbuf_base_off;
    	}
	
    	return 0;
}

/****** Firmware Error Handler **********************************************/

/*============================================================================
 * Firmware error handler.
 *	This routine is called whenever firmware command returns non-zero
 *	return code.
 *
 * Return zero if previous command has to be cancelled.
 */
static int ppp_error(sdla_t *card, int err, wan_mbox_t* mbox)
{
	switch (err){ 
	case WAN_CMD_TIMEOUT:
		DEBUG_EVENT("%s: command 0x%02x timed out!\n", 
				card->devname, mbox->wan_command);
		break;

	default:
		DEBUG_EVENT("%s: command 0x%02x returned 0x%02x!\n",
				card->devname, mbox->wan_command, err);
		break;
	}
	return 0;
}

static void ppp_handle_front_end_state(void* card_id)
{
	sdla_t*		card = (sdla_t*)card_id;

	WAN_ASSERT1(card == NULL);
	/* FIXME: Add handling front end status changes */

	return;
}

/*
 ******************************************************************************
 *			ppp_snmp_data()	
 *
 * Description: Save snmp request and parameters in private structure, enable
 *		TIMER interrupt, put current process in sleep.
 * Arguments:   
 * Returns:
 ******************************************************************************
 */
#define   PPPLINKSTATUSPHYSICALINDEX  			3
#define   PPPLINKSTATUSBADADDRESSES  			4
#define   PPPLINKSTATUSBADCONTROLS  			5
#define   PPPLINKSTATUSPACKETTOOLONGS  			6
#define   PPPLINKSTATUSBADFCSS  			7
#define   PPPLINKSTATUSLOCALMRU  			8
#define   PPPLINKSTATUSREMOTEMRU  			9
#define   PPPLINKSTATUSLOCALTOPEERACCMAP  		10
#define   PPPLINKSTATUSPEERTOLOCALACCMAP  		11
#define   PPPLINKSTATUSLOCALTOREMOTEPROTOCOLCOMPRESSION 12
#define   PPPLINKSTATUSREMOTETOLOCALPROTOCOLCOMPRESSION 13
#define   PPPLINKSTATUSLOCALTOREMOTEACCOMPRESSION  	14
#define   PPPLINKSTATUSREMOTETOLOCALACCOMPRESSION  	15
#define   PPPLINKSTATUSTRANSMITFCSSIZE  		16
#define   PPPLINKSTATUSRECEIVEFCSSIZE  			17
#define   PPPLINKCONFIGINITIALMRU  			20
#define   PPPLINKCONFIGRECEIVEACCMAP  			21
#define   PPPLINKCONFIGTRANSMITACCMAP  			22
#define   PPPLINKCONFIGMAGICNUMBER  			23
#define   PPPLINKCONFIGFCSSIZE  			24
#define   PPPLQRQUALITY         			27
#define   PPPLQRINGOODOCTETS    			28
#define   PPPLQRLOCALPERIOD     			29
#define   PPPLQRREMOTEPERIOD    			30
#define   PPPLQROUTLQRS         			31
#define   PPPLQRINLQRS          			32
#define   PPPLQRCONFIGPERIOD    			35
#define   PPPLQRCONFIGSTATUS    			36
#define   PPPLQREXTNSLASTRECEIVEDLQRPACKET  		39
#define   PPPSECURITYCONFIGLINK  			42
#define   PPPSECURITYCONFIGPREFERENCE  			43
#define   PPPSECURITYCONFIGPROTOCOL  			44
#define   PPPSECURITYCONFIGSTATUS  			45
#define   PPPSECURITYSECRETSLINK  			48
#define   PPPSECURITYSECRETSIDINDEX  			49
#define   PPPSECURITYSECRETSDIRECTION  			50
#define   PPPSECURITYSECRETSPROTOCOL  			51
#define   PPPSECURITYSECRETSIDENTITY  			52
#define   PPPSECURITYSECRETSSECRET  			53
#define   PPPSECURITYSECRETSSTATUS  			54
#define   PPPIPOPERSTATUS       			57
#define   PPPIPLOCALTOREMOTECOMPRESSIONPROTOCOL  	58
#define   PPPIPREMOTETOLOCALCOMPRESSIONPROTOCOL  	59
#define   PPPIPREMOTEMAXSLOTID  			60
#define   PPPIPLOCALMAXSLOTID   			61
#define   PPPIPCONFIGADMINSTATUS  			64
#define   PPPIPCONFIGCOMPRESSION  			65
#define   PPPBRIDGEOPERSTATUS   			68
#define   PPPBRIDGELOCALTOREMOTETINYGRAMCOMPRESSION  	69
#define   PPPBRIDGEREMOTETOLOCALTINYGRAMCOMPRESSION  	70
#define   PPPBRIDGELOCALTOREMOTELANID  			71
#define   PPPBRIDGEREMOTETOLOCALLANID  			72
#define   PPPBRIDGECONFIGADMINSTATUS  			75
#define   PPPBRIDGECONFIGTINYGRAM  			76
#define   PPPBRIDGECONFIGRINGID  			77
#define   PPPBRIDGECONFIGLINEID  			78
#define   PPPBRIDGECONFIGLANID  			79
#define   PPPBRIDGEMEDIAMACTYPE  			82
#define   PPPBRIDGEMEDIALOCALSTATUS  			83
#define   PPPBRIDGEMEDIAREMOTESTATUS  			84
#define   PPPBRIDGEMEDIACONFIGMACTYPE  			87

static int ppp_snmp_data(sdla_t* card, struct ifnet* ifp, void* data)
{
	wanpipe_snmp_t*		snmp;

#if 0
	/* Update device statistics */
	if (card->wandev.update) {
		int rslt = 0;
		card->u.f.update_dlci = chan;
		rslt = card->wandev.update(&card->wandev);
		if(rslt) {
			return (rslt) ? (-EBUSY) : (-EINVAL);
		}
	}
#endif
	snmp = (wanpipe_snmp_t*)data;

	switch(snmp->snmp_magic){
	/* PPP Link Group */
	case PPPLINKSTATUSPHYSICALINDEX: 
	case PPPLINKSTATUSBADADDRESSES: 
	case PPPLINKSTATUSBADCONTROLS: 
            	return -EINVAL;

	case PPPLINKSTATUSPACKETTOOLONGS:
		snmp->snmp_val = 0;
		break;

	case PPPLINKSTATUSBADFCSS:
	case PPPLINKSTATUSLOCALMRU:
	case PPPLINKSTATUSREMOTEMRU:
	case PPPLINKSTATUSLOCALTOPEERACCMAP:
	case PPPLINKSTATUSPEERTOLOCALACCMAP:
	case PPPLINKSTATUSLOCALTOREMOTEPROTOCOLCOMPRESSION:
	case PPPLINKSTATUSREMOTETOLOCALPROTOCOLCOMPRESSION:
	case PPPLINKSTATUSLOCALTOREMOTEACCOMPRESSION:
	case PPPLINKSTATUSREMOTETOLOCALACCOMPRESSION:
	case PPPLINKSTATUSTRANSMITFCSSIZE:
	case PPPLINKSTATUSRECEIVEFCSSIZE:
	case PPPLINKCONFIGINITIALMRU:
	case PPPLINKCONFIGRECEIVEACCMAP:
	case PPPLINKCONFIGTRANSMITACCMAP:
	case PPPLINKCONFIGMAGICNUMBER:
	case PPPLINKCONFIGFCSSIZE:
	case PPPLQRQUALITY:
	case PPPLQRINGOODOCTETS:
	case PPPLQRLOCALPERIOD:
	case PPPLQRREMOTEPERIOD:
	case PPPLQROUTLQRS:
	case PPPLQRINLQRS:
	case PPPLQRCONFIGPERIOD:
	case PPPLQRCONFIGSTATUS:
	case PPPLQREXTNSLASTRECEIVEDLQRPACKET:
	case PPPSECURITYCONFIGLINK:
	case PPPSECURITYCONFIGPREFERENCE:
	case PPPSECURITYCONFIGPROTOCOL:
	case PPPSECURITYCONFIGSTATUS:
	case PPPSECURITYSECRETSLINK:
	case PPPSECURITYSECRETSIDINDEX:
	case PPPSECURITYSECRETSDIRECTION:
	case PPPSECURITYSECRETSPROTOCOL:
	case PPPSECURITYSECRETSIDENTITY:
	case PPPSECURITYSECRETSSECRET:
	case PPPSECURITYSECRETSSTATUS:
	case PPPIPOPERSTATUS:
	case PPPIPLOCALTOREMOTECOMPRESSIONPROTOCOL:
	case PPPIPREMOTETOLOCALCOMPRESSIONPROTOCOL:
	case PPPIPREMOTEMAXSLOTID:
	case PPPIPLOCALMAXSLOTID:
	case PPPIPCONFIGADMINSTATUS:
	case PPPIPCONFIGCOMPRESSION:
	case PPPBRIDGEOPERSTATUS:
	case PPPBRIDGELOCALTOREMOTETINYGRAMCOMPRESSION:
	case PPPBRIDGEREMOTETOLOCALTINYGRAMCOMPRESSION:
	case PPPBRIDGELOCALTOREMOTELANID:
	case PPPBRIDGEREMOTETOLOCALLANID:
	case PPPBRIDGECONFIGADMINSTATUS:
	case PPPBRIDGECONFIGTINYGRAM:
	case PPPBRIDGECONFIGRINGID:
	case PPPBRIDGECONFIGLINEID:
	case PPPBRIDGECONFIGLANID:
	case PPPBRIDGEMEDIAMACTYPE:
	case PPPBRIDGEMEDIALOCALSTATUS:
	case PPPBRIDGEMEDIAREMOTESTATUS:
	case PPPBRIDGEMEDIACONFIGMACTYPE:

	default:
            	return -EINVAL;
	}

	return 0;
}

/*
****************************************************************************
**
**
**
*/
static int ppp_debugging(sdla_t* card)
{
	ppp_pkt_stats_t		pkt_stat;
	ppp_lcp_stats_t		lcp_stat;
	static unsigned long	rx_lcp = 0;
	static unsigned long	rx_conf_nak = 0;

	ppp_get_pkt_stats(card, &pkt_stat);
	ppp_get_lcp_stats(card, &lcp_stat);
	if (card->wan_debugging_state == WAN_DEBUGGING_CONT){
		rx_lcp		= pkt_stat.rx_lcp;
		rx_conf_nak	= lcp_stat.rx_conf_nak;
		card->wan_debugging_state = WAN_DEBUGGING_PROTOCOL;
		return 15;
	}

	if (pkt_stat.rx_lcp != rx_lcp){
		if (lcp_stat.rx_conf_nak != rx_conf_nak){
			if (card->wan_debug_last_msg != WAN_DEBUG_PPP_LCP_MSG){
				DEBUG_EVENT("%s: PPP NAKs received!\n",
							card->devname);
				DEBUG_EVENT("%s:    Check remote router IP configuration!\n",
							card->devname);
				DEBUG_EVENT("%s:    If you cannot resolve this, call your dealer.\n",
						card->devname);	
			}
			card->wan_debug_last_msg = WAN_DEBUG_PPP_LCP_MSG;
		}else{
			if (card->wan_debug_last_msg != WAN_DEBUG_PPP_NAK_MSG){
				DEBUG_EVENT("%s: PPP negotiation problem!\n",
							card->devname);
				DEBUG_EVENT("%s:    Check remote router or ISP.\n",
							card->devname);	
			}
			card->wan_debug_last_msg = WAN_DEBUG_PPP_NAK_MSG;
		}
	}else{
		if (card->wan_debug_last_msg != WAN_DEBUG_PPP_NEG_MSG){
			DEBUG_EVENT("%s: No replies to LCP!\n",
						card->devname);
			DEBUG_EVENT("%s:    Check configuration of remote router.\n",
						card->devname);	
		}
		card->wan_debug_last_msg = WAN_DEBUG_PPP_NEG_MSG;
	}
	
	return 0;
}

static unsigned long ppp_crc_frames(sdla_t* card)
{
	ppp_err_stats_t stats;
	if (!ppp_get_err_stats(card, &stats)){
		return stats.rx_bad_crc;
	}
	return 0;
}

static unsigned long ppp_abort_frames(sdla_t * card)
{
	ppp_err_stats_t stats;
	if (!ppp_get_err_stats(card, &stats)){
		return stats.rx_abort;
	}
	return 0;
}

static unsigned long ppp_tx_underun_frames(sdla_t* card)
{
	ppp_err_stats_t stats;
	if (!ppp_get_err_stats(card, &stats)){
		return stats.tx_missed_intr;
	}
	return 0;
}


/****** End *****************************************************************/
